# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging
import math
import torch
from torch import nn
from torch.nn import functional as F

from minihack.agent.polybeast.models.base import BaseNet, PAD_CHAR, NUM_CHARS
from minihack.agent.common.models.embed import GlyphEmbedding
from minihack.agent.common.models.dynamics import (
    ForwardDynamicsNet,
    InverseDynamicsNet,
)


class IntrinsicRewardNet(BaseNet):
    def __init__(self, observation_shape, num_actions, flags, device):
        super(IntrinsicRewardNet, self).__init__(
            observation_shape, num_actions, flags, device
        )
        self.register_buffer("intrinsic_sum", torch.zeros(()))
        self.register_buffer("intrinsic_m2", torch.zeros(()))
        self.register_buffer("intrinsic_count", torch.zeros(()).fill_(1e-8))

        self.intrinsic_input = flags.int.input

        self.int_baseline = nn.Linear(self.h_dim, 1)

    def intrinsic_enabled(self):
        return True

    @torch.no_grad()
    def update_intrinsic_moments(self, reward_batch):
        """Maintains a running mean of reward."""
        new_count = len(reward_batch)
        new_sum = torch.sum(reward_batch)
        new_mean = new_sum / new_count

        curr_mean = self.intrinsic_sum / self.intrinsic_count
        new_m2 = torch.sum((reward_batch - new_mean) ** 2) + (
            (self.intrinsic_count * new_count)
            / (self.intrinsic_count + new_count)
            * (new_mean - curr_mean) ** 2
        )

        self.intrinsic_count += new_count
        self.intrinsic_sum += new_sum
        self.intrinsic_m2 += new_m2

    @torch.no_grad()
    def get_intrinsic_std(self):
        """Returns standard deviation of the running mean of the intrinsic reward."""
        return torch.sqrt(self.intrinsic_m2 / self.intrinsic_count)


class RNDNet(IntrinsicRewardNet):
    def __init__(self, observation_shape, num_actions, flags, device):
        super(RNDNet, self).__init__(
            observation_shape, num_actions, flags, device
        )

        if self.equalize_input_dim:
            raise NotImplementedError(
                "rnd model does not support equalize_input_dim"
            )

        Y = 8  # number of output filters

        # IMPLEMENTED HERE: RND net using the default feature extractor
        self.rndtgt_embed = GlyphEmbedding(
            flags.glyph_type,
            flags.embedding_dim,
            device,
            flags.use_index_select,
        ).requires_grad_(False)
        self.rndprd_embed = GlyphEmbedding(
            flags.glyph_type,
            flags.embedding_dim,
            device,
            flags.use_index_select,
        )

        if self.intrinsic_input not in ("crop_only", "glyph_only", "full"):
            raise NotImplementedError(
                "RND input type %s" % self.intrinsic_input
            )

        rnd_out_dim = 0
        if self.intrinsic_input in ("crop_only", "full"):
            self.rndtgt_extract_crop_representation = copy.deepcopy(
                self.extract_crop_representation
            ).requires_grad_(False)
            self.rndprd_extract_crop_representation = copy.deepcopy(
                self.extract_crop_representation
            )

            rnd_out_dim += self.crop_dim ** 2 * Y  # crop dim

        if self.intrinsic_input in ("full", "glyph_only"):
            self.rndtgt_extract_representation = copy.deepcopy(
                self.extract_representation
            ).requires_grad_(False)
            self.rndprd_extract_representation = copy.deepcopy(
                self.extract_representation
            )
            rnd_out_dim += self.H * self.W * Y  # glyph dim

            if self.intrinsic_input == "full":
                self.rndtgt_embed_features = nn.Sequential(
                    nn.Linear(self.num_features, self.k_dim),
                    nn.ELU(),
                    nn.Linear(self.k_dim, self.k_dim),
                    nn.ELU(),
                ).requires_grad_(False)
                self.rndprd_embed_features = nn.Sequential(
                    nn.Linear(self.num_features, self.k_dim),
                    nn.ELU(),
                    nn.Linear(self.k_dim, self.k_dim),
                    nn.ELU(),
                )
                rnd_out_dim += self.k_dim  # feature dim

        if self.intrinsic_input == "full" and self.msg_model != "none":
            # we only implement the lt_cnn msg model for RND for simplicity & speed
            if self.msg_model != "lt_cnn":
                logging.warning(
                    "msg.model set to %s, but RND overriding to lt_cnn for its input--"
                    "so the policy and RND are using different models for the messages"
                    % self.msg_model
                )

            self.rndtgt_char_lt = nn.Embedding(
                NUM_CHARS, self.msg_edim, padding_idx=PAD_CHAR
            ).requires_grad_(False)
            self.rndprd_char_lt = nn.Embedding(
                NUM_CHARS, self.msg_edim, padding_idx=PAD_CHAR
            )

            # similar to Zhang et al, 2016
            # Character-level Convolutional Networks for Text Classification
            # https://arxiv.org/abs/1509.01626
            # replace one-hot inputs with learned embeddings
            self.rndtgt_conv1 = nn.Conv1d(
                self.msg_edim, self.msg_hdim, kernel_size=7
            ).requires_grad_(False)
            self.rndprd_conv1 = nn.Conv1d(
                self.msg_edim, self.msg_hdim, kernel_size=7
            )

            # remaining convolutions, relus, pools, and a small FC network
            self.rndtgt_conv2_6_fc = copy.deepcopy(
                self.conv2_6_fc
            ).requires_grad_(False)
            self.rndprd_conv2_6_fc = copy.deepcopy(self.conv2_6_fc)
            rnd_out_dim += self.msg_hdim

        self.rndtgt_fc = (
            nn.Sequential(  # matching RND paper making this smaller
                nn.Linear(rnd_out_dim, self.h_dim)
            ).requires_grad_(False)
        )
        self.rndprd_fc = (
            nn.Sequential(  # matching RND paper making this bigger
                nn.Linear(rnd_out_dim, self.h_dim),
                nn.ELU(),
                nn.Linear(self.h_dim, self.h_dim),
                nn.ELU(),
                nn.Linear(self.h_dim, self.h_dim),
            )
        )

        modules_to_init = [
            self.rndtgt_embed,
            self.rndprd_embed,
            self.rndtgt_fc,
            self.rndprd_fc,
        ]

        SQRT_2 = math.sqrt(2)

        def init(p):
            if isinstance(p, nn.Conv2d) or isinstance(p, nn.Linear):
                # init method used in paper
                nn.init.orthogonal_(p.weight, SQRT_2)
                p.bias.data.zero_()
            if isinstance(p, nn.Embedding):
                nn.init.orthogonal_(p.weight, SQRT_2)

        # manually init all to orthogonal dist

        if self.intrinsic_input in ("full", "crop_only"):
            modules_to_init.append(self.rndtgt_extract_crop_representation)
            modules_to_init.append(self.rndprd_extract_crop_representation)
        if self.intrinsic_input in ("full", "glyph_only"):
            modules_to_init.append(self.rndtgt_extract_representation)
            modules_to_init.append(self.rndprd_extract_representation)
        if self.intrinsic_input == "full":
            modules_to_init.append(self.rndtgt_embed_features)
            modules_to_init.append(self.rndprd_embed_features)
            if self.msg_model != "none":
                modules_to_init.append(self.rndtgt_conv2_6_fc)
                modules_to_init.append(self.rndprd_conv2_6_fc)

        for m in modules_to_init:
            for p in m.modules():
                init(p)

    def forward(self, inputs, core_state, learning=False):
        if not learning:
            # no need to calculate RND outputs when not in learn step
            return super(RNDNet, self).forward(inputs, core_state, learning)
        T, B, *_ = inputs["glyphs"].shape

        glyphs, features = self.prepare_input(inputs)

        # -- [B x 2] x,y coordinates
        coordinates = features[:, :2]

        features = features.view(T * B, -1).float()
        # -- [B x K]
        features_emb = self.embed_features(features)

        assert features_emb.shape[0] == T * B

        reps = [features_emb]

        # -- [B x H' x W']
        crop = self.glyph_embedding.GlyphTuple(
            *[self.crop(g, coordinates) for g in glyphs]
        )
        # -- [B x H' x W' x K]
        crop_emb = self.glyph_embedding(crop)

        if self.crop_model == "transformer":
            # -- [B x W' x H' x K]
            crop_rep = self.extract_crop_representation(crop_emb, mask=None)
        elif self.crop_model == "cnn":
            # -- [B x K x W' x H']
            crop_emb = crop_emb.transpose(1, 3)
            # -- [B x W' x H' x K]
            crop_rep = self.extract_crop_representation(crop_emb)
        # -- [B x K']

        crop_rep = crop_rep.view(T * B, -1)
        assert crop_rep.shape[0] == T * B

        reps.append(crop_rep)

        # -- [B x H x W x K]
        glyphs_emb = self.glyph_embedding(glyphs)
        # -- [B x K x W x H]
        glyphs_emb = glyphs_emb.transpose(1, 3)
        # -- [B x W x H x K]
        glyphs_rep = self.extract_representation(glyphs_emb)

        # -- [B x K']
        glyphs_rep = glyphs_rep.view(T * B, -1)
        if self.equalize_input_dim:
            glyphs_rep = self.project_glyph_dim(glyphs_rep)

        assert glyphs_rep.shape[0] == T * B

        # -- [B x K'']
        reps.append(glyphs_rep)

        # MESSAGING MODEL
        if self.msg_model != "none":
            # [T x B x 256] -> [T * B x 256]
            messages = inputs["message"].long().view(T * B, -1)
            if self.msg_model == "cnn":
                # convert messages to one-hot, [T * B x 96 x 256]
                one_hot = F.one_hot(messages, num_classes=NUM_CHARS).transpose(
                    1, 2
                )
                char_rep = self.conv2_6_fc(self.conv1(one_hot.float()))
            elif self.msg_model == "lt_cnn":
                # [ T * B x E x 256 ]
                char_emb = self.char_lt(messages).transpose(1, 2)
                char_rep = self.conv2_6_fc(self.conv1(char_emb))
            else:  # lstm, gru
                char_emb = self.char_lt(messages)
                output = self.char_rnn(char_emb)[0]
                fwd_rep = output[:, -1, : self.h_dim // 2]
                bwd_rep = output[:, 0, self.h_dim // 2 :]
                char_rep = torch.cat([fwd_rep, bwd_rep], dim=1)

            if self.equalize_input_dim:
                char_rep = self.project_msg_dim(char_rep)
            reps.append(char_rep)

        st = torch.cat(reps, dim=1)

        # -- [B x K]
        st = self.fc(st)

        # TARGET NETWORK
        with torch.no_grad():
            if self.intrinsic_input == "crop_only":
                tgt_crop_emb = self.rndtgt_embed(crop).transpose(1, 3)
                tgt_crop_rep = self.rndtgt_extract_crop_representation(
                    tgt_crop_emb
                )
                tgt_st = self.rndtgt_fc(tgt_crop_rep.view(T * B, -1))
            elif self.intrinsic_input == "glyph_only":
                tgt_glyphs_emb = self.rndtgt_embed(glyphs).transpose(1, 3)
                tgt_glyphs_rep = self.rndtgt_extract_representation(
                    tgt_glyphs_emb
                )
                tgt_st = self.rndtgt_fc(tgt_glyphs_rep.view(T * B, -1))
            else:  # full
                tgt_reps = []
                tgt_feats = self.rndtgt_embed_features(features)
                tgt_reps.append(tgt_feats)

                tgt_crop_emb = self.rndtgt_embed(crop).transpose(1, 3)
                tgt_crop_rep = self.rndtgt_extract_crop_representation(
                    tgt_crop_emb
                )
                tgt_reps.append(tgt_crop_rep.view(T * B, -1))

                tgt_glyphs_emb = self.rndtgt_embed(glyphs).transpose(1, 3)
                tgt_glyphs_rep = self.rndtgt_extract_representation(
                    tgt_glyphs_emb
                )
                tgt_reps.append(tgt_glyphs_rep.view(T * B, -1))

                if self.msg_model != "none":
                    tgt_char_emb = self.rndtgt_char_lt(messages).transpose(
                        1, 2
                    )
                    tgt_char_rep = self.rndtgt_conv2_6_fc(
                        self.rndprd_conv1(tgt_char_emb)
                    )
                    tgt_reps.append(tgt_char_rep)

                tgt_st = self.rndtgt_fc(torch.cat(tgt_reps, dim=1))

        # PREDICTOR NETWORK
        if self.intrinsic_input == "crop_only":
            prd_crop_emb = self.rndprd_embed(crop).transpose(1, 3)
            prd_crop_rep = self.rndprd_extract_crop_representation(
                prd_crop_emb
            )
            prd_st = self.rndprd_fc(prd_crop_rep.view(T * B, -1))
        elif self.intrinsic_input == "glyph_only":
            prd_glyphs_emb = self.rndprd_embed(glyphs).transpose(1, 3)
            prd_glyphs_rep = self.rndprd_extract_representation(prd_glyphs_emb)
            prd_st = self.rndprd_fc(prd_glyphs_rep.view(T * B, -1))
        else:  # full
            prd_reps = []
            prd_feats = self.rndprd_embed_features(features)
            prd_reps.append(prd_feats)

            prd_crop_emb = self.rndprd_embed(crop).transpose(1, 3)
            prd_crop_rep = self.rndprd_extract_crop_representation(
                prd_crop_emb
            )
            prd_reps.append(prd_crop_rep.view(T * B, -1))

            prd_glyphs_emb = self.rndprd_embed(glyphs).transpose(1, 3)
            prd_glyphs_rep = self.rndprd_extract_representation(prd_glyphs_emb)
            prd_reps.append(prd_glyphs_rep.view(T * B, -1))

            if self.msg_model != "none":
                prd_char_emb = self.rndprd_char_lt(messages).transpose(1, 2)
                prd_char_rep = self.rndprd_conv2_6_fc(
                    self.rndprd_conv1(prd_char_emb)
                )
                prd_reps.append(prd_char_rep)

            prd_st = self.rndprd_fc(torch.cat(prd_reps, dim=1))

        assert tgt_st.size() == prd_st.size()

        if self.use_lstm:
            core_input = st.view(T, B, -1)
            core_output_list = []
            notdone = (~inputs["done"]).float()
            for input, nd in zip(core_input.unbind(), notdone.unbind()):
                # Reset core state to zero whenever an episode ended.
                # Make `done` broadcastable with (num_layers, B, hidden_size)
                # states:
                nd = nd.view(1, -1, 1)
                core_state = tuple(nd * t for t in core_state)
                output, core_state = self.core(input.unsqueeze(0), core_state)
                core_output_list.append(output)
            core_output = torch.flatten(torch.cat(core_output_list), 0, 1)
        else:
            core_output = st

        # -- [B x A]
        policy_logits = self.policy(core_output)
        # -- [B x A]
        baseline = self.baseline(core_output)

        if self.training:
            action = torch.multinomial(
                F.softmax(policy_logits, dim=1), num_samples=1
            )
        else:
            # Don't sample when testing.
            action = torch.argmax(policy_logits, dim=1)

        policy_logits = policy_logits.view(T, B, self.num_actions)
        baseline = baseline.view(T, B)
        action = action.view(T, B)

        output = dict(
            policy_logits=policy_logits,
            baseline=baseline,
            action=action,
            target=tgt_st.view(T, B, -1),
            predicted=prd_st.view(T, B, -1),
            int_baseline=self.int_baseline(core_output).view(T, B),
        )
        return (output, core_state)


class RIDENet(IntrinsicRewardNet):
    def __init__(self, observation_shape, num_actions, flags, device):
        super(RIDENet, self).__init__(
            observation_shape, num_actions, flags, device
        )

        if flags.msg.model != "none":
            raise NotImplementedError(
                "model=%s + msg.model=%s" % (flags.model, flags.msg.model)
            )

        self.forward_dynamics_model = ForwardDynamicsNet(
            num_actions,
            flags.ride.hidden_dim,
            flags.hidden_dim,
            flags.hidden_dim,
        )
        self.inverse_dynamics_model = InverseDynamicsNet(
            num_actions,
            flags.ride.hidden_dim,
            flags.hidden_dim,
            flags.hidden_dim,
        )

        Y = 8  # number of output filters

        # IMPLEMENTED HERE: RIDE net using the default feature extractor
        self.ride_embed = GlyphEmbedding(
            flags.glyph_type,
            flags.embedding_dim,
            device,
            flags.use_index_select,
        )

        if self.intrinsic_input not in ("crop_only", "glyph_only", "full"):
            raise NotImplementedError(
                "RIDE input type %s" % self.intrinsic_input
            )

        ride_out_dim = 0
        if self.intrinsic_input in ("crop_only", "full"):
            self.ride_extract_crop_representation = copy.deepcopy(
                self.extract_crop_representation
            )
            ride_out_dim += self.crop_dim ** 2 * Y  # crop dim

        if self.intrinsic_input in ("full", "glyph_only"):
            self.ride_extract_representation = copy.deepcopy(
                self.extract_representation
            )
            ride_out_dim += self.H * self.W * Y  # glyph dim

            if self.intrinsic_input == "full":
                self.ride_embed_features = nn.Sequential(
                    nn.Linear(self.num_features, self.k_dim),
                    nn.ELU(),
                    nn.Linear(self.k_dim, self.k_dim),
                    nn.ELU(),
                )
                ride_out_dim += self.k_dim  # feature dim

        self.ride_fc = nn.Sequential(
            nn.Linear(ride_out_dim, self.h_dim),
            # nn.ELU(),
            # nn.Linear(self.h_dim, self.h_dim),
            # nn.ELU(),
            # nn.Linear(self.h_dim, self.h_dim),
        )

        # reinitialize all deep-copied layers
        modules_to_init = []
        if self.intrinsic_input in ("full", "crop_only"):
            modules_to_init.append(self.ride_extract_crop_representation)
        if self.intrinsic_input in ("full", "glyph_only"):
            modules_to_init.append(self.ride_extract_representation)

        for m in modules_to_init:
            for p in m.modules():
                if isinstance(p, nn.Conv2d):
                    p.reset_parameters()

    def forward(self, inputs, core_state, learning=False):
        if not learning:
            # no need to calculate RIDE outputs when not in learn step
            return super(RIDENet, self).forward(inputs, core_state, learning)

        T, B, *_ = inputs["glyphs"].shape

        glyphs, features = self.prepare_input(inputs)

        # -- [B x 2] x,y coordinates
        coordinates = features[:, :2]

        features = features.view(T * B, -1).float()
        # -- [B x K]
        features_emb = self.embed_features(features)

        assert features_emb.shape[0] == T * B

        reps = [features_emb]

        # -- [B x H' x W']
        crop = self.glyph_embedding.GlyphTuple(
            *[self.crop(g, coordinates) for g in glyphs]
        )
        # -- [B x H' x W' x K]
        crop_emb = self.glyph_embedding(crop)

        if self.crop_model == "transformer":
            # -- [B x W' x H' x K]
            crop_rep = self.extract_crop_representation(crop_emb, mask=None)
        elif self.crop_model == "cnn":
            # -- [B x K x W' x H']
            crop_emb = crop_emb.transpose(1, 3)
            # -- [B x W' x H' x K]
            crop_rep = self.extract_crop_representation(crop_emb)
        # -- [B x K']

        crop_rep = crop_rep.view(T * B, -1)
        assert crop_rep.shape[0] == T * B

        reps.append(crop_rep)

        # -- [B x H x W x K]
        glyphs_emb = self.glyph_embedding(glyphs)
        # glyphs_emb = self.embed(glyphs)
        # -- [B x K x W x H]
        glyphs_emb = glyphs_emb.transpose(1, 3)
        # -- [B x W x H x K]
        glyphs_rep = self.extract_representation(glyphs_emb)

        # -- [B x K']
        glyphs_rep = glyphs_rep.view(T * B, -1)
        assert glyphs_rep.shape[0] == T * B

        # -- [B x K'']
        reps.append(glyphs_rep)

        st = torch.cat(reps, dim=1)

        # -- [B x K]
        st = self.fc(st)

        # PREDICTOR NETWORK
        if self.intrinsic_input == "crop_only":
            ride_crop_emb = self.ride_embed(crop).transpose(1, 3)
            ride_crop_rep = self.ride_extract_crop_representation(
                ride_crop_emb
            )
            ride_st = self.ride_fc(ride_crop_rep.view(T * B, -1))
        elif self.intrinsic_input == "glyph_only":
            ride_glyphs_emb = self.ride_embed(glyphs).transpose(1, 3)
            ride_glyphs_rep = self.ride_extract_representation(ride_glyphs_emb)
            ride_st = self.ride_fc(ride_glyphs_rep.view(T * B, -1))
        else:  # full
            ride_reps = []
            ride_feats = self.ride_embed_features(features)
            ride_reps.append(ride_feats)

            ride_crop_emb = self.ride_embed(crop).transpose(1, 3)
            ride_crop_rep = self.ride_extract_crop_representation(
                ride_crop_emb
            )
            ride_reps.append(ride_crop_rep.view(T * B, -1))

            ride_glyphs_emb = self.ride_embed(glyphs).transpose(1, 3)
            ride_glyphs_rep = self.ride_extract_representation(ride_glyphs_emb)
            ride_reps.append(ride_glyphs_rep.view(T * B, -1))

            ride_st = self.ride_fc(torch.cat(ride_reps, dim=1))

        if self.use_lstm:
            core_input = st.view(T, B, -1)
            core_output_list = []
            notdone = (~inputs["done"]).float()
            for input, nd in zip(core_input.unbind(), notdone.unbind()):
                # Reset core state to zero whenever an episode ended.
                # Make `done` broadcastable with (num_layers, B, hidden_size)
                # states:
                nd = nd.view(1, -1, 1)
                core_state = tuple(nd * t for t in core_state)
                output, core_state = self.core(input.unsqueeze(0), core_state)
                core_output_list.append(output)
            core_output = torch.flatten(torch.cat(core_output_list), 0, 1)
        else:
            core_output = st

        # -- [B x A]
        policy_logits = self.policy(core_output)
        # -- [B x A]
        baseline = self.baseline(core_output)

        if self.training:
            action = torch.multinomial(
                F.softmax(policy_logits, dim=1), num_samples=1
            )
        else:
            # Don't sample when testing.
            action = torch.argmax(policy_logits, dim=1)

        policy_logits = policy_logits.view(T, B, self.num_actions)
        baseline = baseline.view(T, B)
        action = action.view(T, B)

        output = dict(
            policy_logits=policy_logits,
            baseline=baseline,
            action=action,
            state_embedding=ride_st.view(T, B, -1),
            int_baseline=self.int_baseline(core_output).view(T, B),
        )
        return (output, core_state)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from minihack.agent.common.envs import tasks
from minihack.agent.polybeast.models.base import BaseNet, RandomNet
from minihack.agent.polybeast.models.intrinsic import RNDNet, RIDENet
from nle.env.base import DUNGEON_SHAPE


def create_model(flags, device):
    model_string = flags.model
    if model_string == "random":
        model_cls = RandomNet
    elif model_string == "baseline":
        model_cls = BaseNet
    elif model_string == "rnd":
        model_cls = RNDNet
    elif model_string == "ride":
        model_cls = RIDENet
    elif model_string == "cnn" or model_string == "transformer":
        raise RuntimeError(
            "model=%s deprecated, use model=baseline crop_model=%s instead"
            % (model_string, model_string)
        )
    else:
        raise NotImplementedError("model=%s" % model_string)

    num_actions = len(
        tasks.ENVS[flags.env](savedir=None, archivefile=None)._actions
    )

    model = model_cls(DUNGEON_SHAPE, num_actions, flags, device)
    model.to(device=device)
    return model



# Copyright (c) Facebook, Inc. and its affiliates.
from minihack.envs import register
from minihack import MiniHackNavigation
from minihack.envs.corridor import NAVIGATE_ACTIONS
from minihack.reward_manager import RewardManager
from nle.nethack import Command

EAT_ACTION = Command.EAT
ACTIONS = tuple(list(NAVIGATE_ACTIONS) + [EAT_ACTION])


def stairs_reward_function(env, previous_observation, action, observation):
    # Agent is on stairs down
    if observation[env._internal_index][4]:
        return 1
    return 0


class MiniHackExploreMaze(MiniHackNavigation):
    """Environment for a memory challenge."""

    def __init__(self, *args, des_file, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 500)
        kwargs["actions"] = ACTIONS
        kwargs["autopickup"] = False
        kwargs["allow_all_yn_questions"] = True
        reward_manager = RewardManager()
        reward_manager.add_eat_event(
            "apple",
            reward=0.5,
            repeatable=True,
            terminal_required=False,
            terminal_sufficient=False,
        )
        # Will never be achieved, but insures the environment keeps running
        reward_manager.add_message_event(
            ["Mission Complete."],
            terminal_required=True,
            terminal_sufficient=True,
        )
        reward_manager.add_custom_reward_fn(stairs_reward_function)
        super().__init__(
            *args,
            des_file=des_file,
            reward_manager=reward_manager,
            **kwargs,
        )


class MiniHackExploreMazeEasy(MiniHackExploreMaze):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, des_file="exploremazeeasy.des", **kwargs)


class MiniHackExploreMazeHard(MiniHackExploreMaze):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, des_file="exploremazehard.des", **kwargs)


class MiniHackExploreMazeEasyMapped(MiniHackExploreMaze):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, des_file="exploremazeeasy_premapped.des", **kwargs
        )


class MiniHackExploreMazeHardMapped(MiniHackExploreMaze):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, des_file="exploremazehard_premapped.des", **kwargs
        )


register(
    id="MiniHack-ExploreMaze-Easy-v0",
    entry_point="minihack.envs.exploremaze:MiniHackExploreMazeEasy",
)
register(
    id="MiniHack-ExploreMaze-Hard-v0",
    entry_point="minihack.envs.exploremaze:MiniHackExploreMazeHard",
)
register(
    id="MiniHack-ExploreMaze-Easy-Mapped-v0",
    entry_point="minihack.envs.exploremaze:MiniHackExploreMazeEasyMapped",
)
register(
    id="MiniHack-ExploreMaze-Hard-Mapped-v0",
    entry_point="minihack.envs.exploremaze:MiniHackExploreMazeHardMapped",
)



# Copyright (c) Facebook, Inc. and its affiliates.

from minihack.tiles.tile import glyph2tile, MAXMONTILE, MAXOBJTILE, MAXOTHTILE
from minihack.tiles.glyph_mapper import GlyphMapper


__all__ = [GlyphMapper, glyph2tile, MAXMONTILE, MAXOBJTILE, MAXOTHTILE]



# Copyright (c) Facebook, Inc. and its affiliates.
from minihack import MiniHackNavigation
from minihack.level_generator import LevelGenerator
from minihack.envs import register

DUNGEON_SHAPE = (76, 21)


class MiniHackMazeWalk(MiniHackNavigation):
    """Environment for "mazewalk" task."""

    def __init__(self, *args, w, h, premapped, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 1000)
        if premapped:
            flags = ("hardfloor", "premapped")
        else:
            flags = ("hardfloor",)

        if (w, h) != DUNGEON_SHAPE:
            # Fill the level with concrete walls " " surrounded by regular walls
            w, h = w + 2, h + 2
            lvl_gen = LevelGenerator(w=w, h=h, flags=flags)
            lvl_gen.fill_terrain("rect", "-", 0, 0, w - 1, h - 1)
            lvl_gen.fill_terrain("fillrect", " ", 1, 1, w - 2, h - 2)
        else:
            lvl_gen = LevelGenerator(w=w, h=h, fill=" ", flags=flags)

        lvl_gen.add_mazewalk()
        lvl_gen.add_stair_down()

        super().__init__(*args, des_file=lvl_gen.get_des(), **kwargs)


class MiniHackMazeWalk9x9(MiniHackMazeWalk):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 200)
        super().__init__(*args, w=9, h=9, premapped=False, **kwargs)


class MiniHackMazeWalk9x9Premapped(MiniHackMazeWalk):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 200)
        super().__init__(*args, w=9, h=9, premapped=True, **kwargs)


class MiniHackMazeWalk15x15(MiniHackMazeWalk):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, w=15, h=15, premapped=False, **kwargs)


class MiniHackMazeWalk15x15Premapped(MiniHackMazeWalk):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, w=15, h=15, premapped=True, **kwargs)


class MiniHackMazeWalk45x19(MiniHackMazeWalk):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, w=45, h=19, premapped=False, **kwargs)


class MiniHackMazeWalk45x19Premapped(MiniHackMazeWalk):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, w=45, h=19, premapped=True, **kwargs)


register(
    id="MiniHack-MazeWalk-9x9-v0",
    entry_point="minihack.envs.mazewalk:MiniHackMazeWalk9x9",
)
register(
    id="MiniHack-MazeWalk-Mapped-9x9-v0",
    entry_point="minihack.envs.mazewalk:MiniHackMazeWalk9x9Premapped",
)
register(
    id="MiniHack-MazeWalk-15x15-v0",
    entry_point="minihack.envs.mazewalk:MiniHackMazeWalk15x15",
)
register(
    id="MiniHack-MazeWalk-Mapped-15x15-v0",
    entry_point="minihack.envs.mazewalk:MiniHackMazeWalk15x15Premapped",
)
register(
    id="MiniHack-MazeWalk-45x19-v0",
    entry_point="minihack.envs.mazewalk:MiniHackMazeWalk45x19",
)
register(
    id="MiniHack-MazeWalk-Mapped-45x19-v0",
    entry_point="minihack.envs.mazewalk:MiniHackMazeWalk45x19Premapped",
)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum

import numpy as np

from nle.nethack import *  # noqa: F403

# flake8: noqa: F405

NUM_OBJECTS = 453
MAXEXPCHARS = 9


class GlyphGroup(enum.IntEnum):
    # See display.h in NetHack.
    MON = 0
    PET = 1
    INVIS = 2
    DETECT = 3
    BODY = 4
    RIDDEN = 5
    OBJ = 6
    CMAP = 7
    EXPLODE = 8
    ZAP = 9
    SWALLOW = 10
    WARNING = 11
    STATUE = 12


def id_pairs_table():
    """Returns a lookup table for glyph -> NLE id pairs."""
    table = np.zeros([MAX_GLYPH, 2], dtype=np.int16)

    num_nle_ids = 0

    for glyph in range(GLYPH_MON_OFF, GLYPH_PET_OFF):
        table[glyph] = (glyph, GlyphGroup.MON)
        num_nle_ids += 1

    for glyph in range(GLYPH_PET_OFF, GLYPH_INVIS_OFF):
        table[glyph] = (glyph - GLYPH_PET_OFF, GlyphGroup.PET)

    for glyph in range(GLYPH_INVIS_OFF, GLYPH_DETECT_OFF):
        table[glyph] = (num_nle_ids, GlyphGroup.INVIS)
        num_nle_ids += 1

    for glyph in range(GLYPH_DETECT_OFF, GLYPH_BODY_OFF):
        table[glyph] = (glyph - GLYPH_DETECT_OFF, GlyphGroup.DETECT)

    for glyph in range(GLYPH_BODY_OFF, GLYPH_RIDDEN_OFF):
        table[glyph] = (glyph - GLYPH_BODY_OFF, GlyphGroup.BODY)

    for glyph in range(GLYPH_RIDDEN_OFF, GLYPH_OBJ_OFF):
        table[glyph] = (glyph - GLYPH_RIDDEN_OFF, GlyphGroup.RIDDEN)

    for glyph in range(GLYPH_OBJ_OFF, GLYPH_CMAP_OFF):
        table[glyph] = (num_nle_ids, GlyphGroup.OBJ)
        num_nle_ids += 1

    for glyph in range(GLYPH_CMAP_OFF, GLYPH_EXPLODE_OFF):
        table[glyph] = (num_nle_ids, GlyphGroup.CMAP)
        num_nle_ids += 1

    for glyph in range(GLYPH_EXPLODE_OFF, GLYPH_ZAP_OFF):
        id_ = num_nle_ids + (glyph - GLYPH_EXPLODE_OFF) // MAXEXPCHARS
        table[glyph] = (id_, GlyphGroup.EXPLODE)

    num_nle_ids += EXPL_MAX

    for glyph in range(GLYPH_ZAP_OFF, GLYPH_SWALLOW_OFF):
        id_ = num_nle_ids + (glyph - GLYPH_ZAP_OFF) // 4
        table[glyph] = (id_, GlyphGroup.ZAP)

    num_nle_ids += NUM_ZAP

    for glyph in range(GLYPH_SWALLOW_OFF, GLYPH_WARNING_OFF):
        table[glyph] = (num_nle_ids, GlyphGroup.SWALLOW)
    num_nle_ids += 1

    for glyph in range(GLYPH_WARNING_OFF, GLYPH_STATUE_OFF):
        table[glyph] = (num_nle_ids, GlyphGroup.WARNING)
        num_nle_ids += 1

    for glyph in range(GLYPH_STATUE_OFF, MAX_GLYPH):
        table[glyph] = (glyph - GLYPH_STATUE_OFF, GlyphGroup.STATUE)

    return table


def id_pairs_func(glyph):
    result = glyph_to_mon(glyph)
    if result != NO_GLYPH:
        return result
    if glyph_is_invisible(glyph):
        return NUMMONS
    if glyph_is_body(glyph):
        return glyph - GLYPH_BODY_OFF

    offset = NUMMONS + 1

    # CORPSE handled by glyph_is_body; STATUE handled by glyph_to_mon.
    result = glyph_to_obj(glyph)
    if result != NO_GLYPH:
        return result + offset
    offset += NUM_OBJECTS

    # I don't understand glyph_to_cmap and/or the GLYPH_EXPLODE_OFF definition
    # with MAXPCHARS - MAXEXPCHARS.
    if GLYPH_CMAP_OFF <= glyph < GLYPH_EXPLODE_OFF:
        return glyph - GLYPH_CMAP_OFF + offset
    offset += MAXPCHARS - MAXEXPCHARS

    if GLYPH_EXPLODE_OFF <= glyph < GLYPH_ZAP_OFF:
        return (glyph - GLYPH_EXPLODE_OFF) // MAXEXPCHARS + offset
    offset += EXPL_MAX

    if GLYPH_ZAP_OFF <= glyph < GLYPH_SWALLOW_OFF:
        return ((glyph - GLYPH_ZAP_OFF) >> 2) + offset
    offset += NUM_ZAP

    if GLYPH_SWALLOW_OFF <= glyph < GLYPH_WARNING_OFF:
        return offset
    offset += 1

    result = glyph_to_warning(glyph)
    if result != NO_GLYPH:
        return result + offset



# Copyright (c) Facebook, Inc. and its affiliates.

from minihack.tiles import glyph2tile, MAXOTHTILE
from nle.nethack import MAX_GLYPH
import numpy as np
import pkg_resources
import pickle
import os


class GlyphMapper:
    """This class is used to map glyphs to rgb pixels."""

    def __init__(self):
        self.tiles = self.load_tiles()

    def load_tiles(self):
        """This function expects that tile.npy already exists.
        If it doesn't, call make_tiles.py in win/
        """

        tile_rgb_path = os.path.join(
            pkg_resources.resource_filename("minihack", "tiles"),
            "tiles.pkl",
        )

        return pickle.load(open(tile_rgb_path, "rb"))

    def glyph_id_to_rgb(self, glyph_id):
        tile_id = glyph2tile[glyph_id]
        assert 0 <= tile_id <= MAXOTHTILE
        return self.tiles[tile_id]

    def _glyph_to_rgb(self, glyphs):
        # Expects glhyphs as two-dimensional numpy ndarray
        cols = None
        col = None

        for i in range(glyphs.shape[1]):
            for j in range(glyphs.shape[0]):
                rgb = self.glyph_id_to_rgb(glyphs[j, i])
                if col is None:
                    col = rgb
                else:
                    col = np.concatenate((col, rgb))

            if cols is None:
                cols = col
            else:
                cols = np.concatenate((cols, col), axis=1)
            col = None

        return cols

    def to_rgb(self, glyphs):
        return self._glyph_to_rgb(glyphs)



# Copyright (c) Facebook, Inc. and its affiliates.
from minihack.envs import register
from minihack import MiniHackNavigation, RewardManager


class MiniHackMemento(MiniHackNavigation):
    """Environment for a memento challenge."""

    def __init__(self, *args, des_file, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 5000)
        reward_manager = RewardManager()
        reward_manager.add_message_event(
            ["squeak"],
            reward=-1,
            terminal_sufficient=True,
            terminal_required=True,
        )
        reward_manager.add_kill_event(
            "grid bug",
            reward=1,
            terminal_sufficient=True,
            terminal_required=True,
        )
        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackMementoShortF2(MiniHackMemento):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, des_file="memento_short.des", **kwargs)


class MiniHackMementoF2(MiniHackMemento):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, des_file="memento_easy.des", **kwargs)


class MiniHackMementoF4(MiniHackMemento):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, des_file="memento_hard.des", **kwargs)


register(
    id="MiniHack-Memento-Short-F2-v0",
    entry_point="minihack.envs.memento:MiniHackMementoShortF2",
)


register(
    id="MiniHack-Memento-F2-v0",
    entry_point="minihack.envs.memento:MiniHackMementoF2",
)

register(
    id="MiniHack-Memento-F4-v0",
    entry_point="minihack.envs.memento:MiniHackMementoF4",
)



# Copyright (c) Facebook, Inc. and its affiliates.

import os
import zipfile
import pkg_resources

DESTINATION_PATH = pkg_resources.resource_filename("minihack", "dat")
BOXOBAN_REPO_URL = (
    "https://github.com/deepmind/boxoban-levels/archive/refs/heads/master.zip"
)


def download_boxoban_levels():
    print("Downloading Boxoban levels...")
    os.system(
        f"wget -c --read-timeout=5 --tries=0 "
        f'"{BOXOBAN_REPO_URL}" -P {DESTINATION_PATH}'
    )
    print("Boxoban levels downloaded, unpacking...")

    zip_file = os.path.join(DESTINATION_PATH, "master.zip")
    with zipfile.ZipFile(zip_file, "r") as zip_ref:
        zip_ref.extractall(DESTINATION_PATH)

    os.remove(zip_file)


if __name__ == "__main__":
    download_boxoban_levels()



# Copyright (c) Facebook, Inc. and its affiliates.
from minihack import MiniHackSkill, LevelGenerator, RewardManager
from minihack.envs import register

freeze_msgs = [
    "The bolt of cold bounces!",  # checks if cold bounces from the wall
]


class MiniHackFreeze(MiniHackSkill):
    def __init__(self, *args, des_file, **kwargs):
        rwrd_mngr = RewardManager()
        rwrd_mngr.add_message_event(freeze_msgs)

        super().__init__(
            *args, des_file=des_file, reward_manager=rwrd_mngr, **kwargs
        )


class MiniHackFreezeWand(MiniHackFreeze):
    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=8, h=8, lit=True)
        lvl_gen.add_object("cold", "/", cursestate="blessed")
        des_file = lvl_gen.get_des()

        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackFreezeHorn(MiniHackFreeze):
    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=8, h=8, lit=True)
        lvl_gen.add_object("frost horn", "(", cursestate="blessed")
        des_file = lvl_gen.get_des()

        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackFreezeRandom(MiniHackFreeze):
    def __init__(self, *args, **kwargs):
        des_file = """
MAZE: "mylevel", ' '
FLAGS:hardfloor
MESSAGE: "Welcome to MiniHack!"
INIT_MAP: solidfill,' '
GEOMETRY:center,center
MAP
........
........
........
........
........
........
........
........
ENDMAP
REGION:(0,0,7,7),lit,"ordinary"
IF [50%] {
    OBJECT:('/',"cold"),random,blessed
} ELSE {
    OBJECT:('(',"frost horn"),random,blessed
}
"""
        super().__init__(*args, des_file=des_file, **kwargs)


class MiniHackFreezeLava(MiniHackSkill):
    def __init__(self, *args, **kwargs):
        des_file = """
MAZE: "mylevel", ' '
FLAGS:hardfloor
INIT_MAP: solidfill,' '
GEOMETRY:center,center
MAP
-------------
|.....L.....|
|.....L.....|
|.....L.....|
|.....L.....|
|.....L.....|
-------------
ENDMAP
REGION:(0,0,12,6),lit,"ordinary"
$left_bank = selection:fillrect (1,1,5,5)
$right_bank = selection:fillrect (7,1,11,5)
IF [50%] {
    # wand of cold
    OBJECT:('/',"cold"),rndcoord($left_bank),blessed
} ELSE {
    # frost horn
    OBJECT:('(',"frost horn"),rndcoord($left_bank),blessed
}
BRANCH:(1,1,5,5),(0,0,0,0)
STAIR:rndcoord($right_bank),down
"""
        super().__init__(*args, des_file=des_file, **kwargs)


register(
    id="MiniHack-Freeze-Wand-v0",
    entry_point="minihack.envs.skills_freeze:MiniHackFreezeWand",
)
register(
    id="MiniHack-Freeze-Horn-v0",
    entry_point="minihack.envs.skills_freeze:MiniHackFreezeHorn",
)
register(
    id="MiniHack-Freeze-Random-v0",
    entry_point="minihack.envs.skills_freeze:MiniHackFreezeRandom",
)
register(
    id="MiniHack-Freeze-Lava-v0",
    entry_point="minihack.envs.skills_freeze:MiniHackFreezeLava",
)



# Copyright (c) Facebook, Inc. and its affiliates.
from minihack import MiniHackNavigation, LevelGenerator
from nle.nethack import Command, CompassDirection
from minihack.envs import register
import gym


MOVE_AND_KICK_ACTIONS = tuple(
    list(CompassDirection) + [Command.OPEN, Command.KICK]
)


class MiniGridHack(MiniHackNavigation):
    def __init__(self, *args, **kwargs):
        # Only ask users to install gym-minigrid if they actually need it
        try:
            import gym_minigrid  # noqa: F401
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "To use MiniGrid-based environments, please install"
                " gym-minigrid: pip3 install gym-minigrid"
            )

        self.minigrid_env = gym.make(kwargs.pop("env_name"))
        self.num_mon = kwargs.pop("num_mon", 0)
        self.num_trap = kwargs.pop("num_trap", 0)
        self.door_state = kwargs.pop("door_state", "closed")
        if self.door_state == "locked":
            kwargs["actions"] = MOVE_AND_KICK_ACTIONS

        lava_walls = kwargs.pop("lava_walls", False)
        if lava_walls:
            self.wall = "L"
        else:
            self.wall = "|"

        des_file = self.get_env_desc()
        super().__init__(*args, des_file=des_file, **kwargs)

    def get_env_map(self, env):
        door_pos = []
        goal_pos = None
        empty_strs = 0
        empty_str = True
        env_map = []

        for j in range(env.grid.height):
            str = ""
            for i in range(env.width):
                c = env.grid.get(i, j)
                if c is None:
                    str += "."
                    continue
                empty_str = False
                if c.type == "wall":
                    str += self.wall
                elif c.type == "door":
                    str += "+"
                    door_pos.append((i, j - empty_strs))
                elif c.type == "floor":
                    str += "."
                elif c.type == "lava":
                    str += "L"
                elif c.type == "goal":
                    goal_pos = (i, j - empty_strs)
                    str += "."
                elif c.type == "player":
                    str += "."
            if not empty_str and j < env.grid.height - 1:
                if set(str) != {"."}:
                    str = str.replace(".", " ", str.index(self.wall))
                    inv = str[::-1]
                    str = inv.replace(".", " ", inv.index(self.wall))[::-1]
                    env_map.append(str)
            elif empty_str:
                empty_strs += 1

        start_pos = (int(env.agent_pos[0]), int(env.agent_pos[1]) - empty_strs)
        env_map = "\n".join(env_map)

        return env_map, start_pos, goal_pos, door_pos

    def get_env_desc(self):
        self.minigrid_env.reset()
        env = self.minigrid_env

        map, start_pos, goal_pos, door_pos = self.get_env_map(env)

        lev_gen = LevelGenerator(map=map)

        lev_gen.add_goal_pos(goal_pos)
        lev_gen.set_start_pos(start_pos)

        for d in door_pos:
            lev_gen.add_door(self.door_state, d)

        lev_gen.wallify()

        for _ in range(self.num_mon):
            lev_gen.add_monster()

        for _ in range(self.num_trap):
            lev_gen.add_trap()

        return lev_gen.get_des()

    def seed(self, core=None, disp=None, reseed=False):
        """The signature of this method corresponds to that of NLE base class.
        For more information see
        https://github.com/facebookresearch/nle/blob/main/nle/env/base.py.

        Sets the state of the NetHack RNGs after the next reset.

        Arguments:
            core [int or None]: Seed for the core RNG. If None, chose a random
                value.
            disp [int or None]: Seed for the disp (anti-TAS) RNG. If None, chose
                a random value.
            reseed [boolean]: As an Anti-TAS (automation) measure,
                NetHack 3.6 reseeds with true randomness every now and then. This
                flag enables or disables this behavior. If set to True,
                trajectories won't be reproducible.

        Returns:
            [tuple] The seeds supplied, in the form (core, disp, reseed).
        """
        self.minigrid_env.seed(core)
        return super().seed(core, disp, reseed)

    def reset(self, wizkit_items=None):
        des_file = self.get_env_desc()
        self.update(des_file)
        return super().reset(wizkit_items=wizkit_items)


class MiniHackMultiRoomN2(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 40)
        super().__init__(
            *args, env_name="MiniGrid-MultiRoom-N2-S4-v0", **kwargs
        )


class MiniHackMultiRoomN4(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 120)
        super().__init__(
            *args, env_name="MiniGrid-MultiRoom-N4-S5-v0", **kwargs
        )


class MiniHackMultiRoomN6(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 240)
        super().__init__(*args, env_name="MiniGrid-MultiRoom-N6-v0", **kwargs)


register(
    id="MiniHack-MultiRoom-N2-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN2",
)
register(
    id="MiniHack-MultiRoom-N4-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN4",
)
register(
    id="MiniHack-MultiRoom-N6-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN6",
)


# MiniGrid: LockedMultiRoom
class MiniHackMultiRoomN2Locked(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 40)
        kwargs["door_state"] = "locked"
        super().__init__(
            *args, env_name="MiniGrid-MultiRoom-N2-S4-v0", **kwargs
        )


class MiniHackMultiRoomN4Locked(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 120)
        kwargs["door_state"] = "locked"
        super().__init__(
            *args, env_name="MiniGrid-MultiRoom-N4-S5-v0", **kwargs
        )


class MiniHackMultiRoomN6Locked(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 240)
        kwargs["door_state"] = "locked"
        super().__init__(*args, env_name="MiniGrid-MultiRoom-N6-v0", **kwargs)


register(
    id="MiniHack-MultiRoom-N2-Locked-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN2Locked",
)
register(
    id="MiniHack-MultiRoom-N4-Locked-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN4Locked",
)
register(
    id="MiniHack-MultiRoom-N6-Locked-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN6Locked",
)


# MiniGrid: LavaMultiRoom
class MiniHackMultiRoomN2Lava(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 40)
        kwargs["lava_walls"] = True
        super().__init__(
            *args, env_name="MiniGrid-MultiRoom-N2-S4-v0", **kwargs
        )


class MiniHackMultiRoomN4Lava(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 120)
        kwargs["lava_walls"] = True
        super().__init__(
            *args, env_name="MiniGrid-MultiRoom-N4-S5-v0", **kwargs
        )


class MiniHackMultiRoomN6Lava(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 240)
        kwargs["lava_walls"] = True
        super().__init__(*args, env_name="MiniGrid-MultiRoom-N6-v0", **kwargs)


register(
    id="MiniHack-MultiRoom-N2-Lava-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN2Lava",
)
register(
    id="MiniHack-MultiRoom-N4-Lava-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN4Lava",
)
register(
    id="MiniHack-MultiRoom-N6-Lava-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN6Lava",
)


# MiniGrid: MonsterpedMultiRoom
class MiniHackMultiRoomN2Monster(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 40)
        kwargs["num_mon"] = 3
        super().__init__(
            *args, env_name="MiniGrid-MultiRoom-N2-S4-v0", **kwargs
        )


class MiniHackMultiRoomN4Monster(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 120)
        kwargs["num_mon"] = 6
        super().__init__(
            *args, env_name="MiniGrid-MultiRoom-N4-S5-v0", **kwargs
        )


class MiniHackMultiRoomN6Monster(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 240)
        kwargs["num_mon"] = 9
        super().__init__(*args, env_name="MiniGrid-MultiRoom-N6-v0", **kwargs)


# MiniGrid: MonsterMultiRoom
register(
    id="MiniHack-MultiRoom-N2-Monster-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN2Monster",
)
register(
    id="MiniHack-MultiRoom-N4-Monster-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN4Monster",
)
register(
    id="MiniHack-MultiRoom-N6-Monster-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN6Monster",
)


# MiniGrid: ExtremeMultiRoom
class MiniHackMultiRoomN2Extreme(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 40)
        kwargs["num_mon"] = 3
        kwargs["lava_walls"] = True
        kwargs["door_state"] = "locked"
        super().__init__(
            *args, env_name="MiniGrid-MultiRoom-N2-S4-v0", **kwargs
        )


class MiniHackMultiRoomN4Extreme(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 120)
        kwargs["num_mon"] = 6
        kwargs["lava_walls"] = True
        kwargs["door_state"] = "locked"
        super().__init__(
            *args, env_name="MiniGrid-MultiRoom-N4-S5-v0", **kwargs
        )


class MiniHackMultiRoomN6Extreme(MiniGridHack):
    def __init__(self, *args, **kwargs):
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 240)
        kwargs["num_mon"] = 9
        kwargs["lava_walls"] = True
        kwargs["door_state"] = "locked"
        super().__init__(*args, env_name="MiniGrid-MultiRoom-N6-v0", **kwargs)


register(
    id="MiniHack-MultiRoom-N2-Extreme-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN2Extreme",
)
register(
    id="MiniHack-MultiRoom-N4-Extreme-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN4Extreme",
)
register(
    id="MiniHack-MultiRoom-N6-Extreme-v0",
    entry_point="minihack.envs.minigrid:MiniHackMultiRoomN6Extreme",
)

# MiniGrid: LavaCrossing
register(
    id="MiniHack-LavaCrossingS9N1-v0",
    entry_point="minihack.envs.minigrid:MiniGridHack",
    kwargs={"env_name": "MiniGrid-LavaCrossingS9N1-v0"},
)
register(
    id="MiniHack-LavaCrossingS9N2-v0",
    entry_point="minihack.envs.minigrid:MiniGridHack",
    kwargs={"env_name": "MiniGrid-LavaCrossingS9N2-v0"},
)
register(
    id="MiniHack-LavaCrossingS9N3-v0",
    entry_point="minihack.envs.minigrid:MiniGridHack",
    kwargs={"env_name": "MiniGrid-LavaCrossingS9N3-v0"},
)
register(
    id="MiniHack-LavaCrossingS11N5-v0",
    entry_point="minihack.envs.minigrid:MiniGridHack",
    kwargs={"env_name": "MiniGrid-LavaCrossingS11N5-v0"},
)

# MiniGrid: Simple Crossing
register(
    id="MiniHack-SimpleCrossingS9N1-v0",
    entry_point="minihack.envs.minigrid:MiniGridHack",
    kwargs={"env_name": "MiniGrid-SimpleCrossingS9N1-v0"},
)
register(
    id="MiniHack-SimpleCrossingS9N2-v0",
    entry_point="minihack.envs.minigrid:MiniGridHack",
    kwargs={"env_name": "MiniGrid-SimpleCrossingS9N2-v0"},
)
register(
    id="MiniHack-SimpleCrossingS9N3-v0",
    entry_point="minihack.envs.minigrid:MiniGridHack",
    kwargs={"env_name": "MiniGrid-SimpleCrossingS9N3-v0"},
)
register(
    id="MiniHack-SimpleCrossingS11N5-v0",
    entry_point="minihack.envs.minigrid:MiniGridHack",
    kwargs={"env_name": "MiniGrid-SimpleCrossingS11N5-v0"},
)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import multiprocessing as mp
import logging
import os
import time

from minihack.agent.common.envs import create_env
import libtorchbeast


# yapf: disable
parser = argparse.ArgumentParser(description='Remote Environment Server')

parser.add_argument('--env', default='staircase', type=str, metavar='E',
                    help='Name of Gym environment to create.')
parser.add_argument('--character', default='mon-hum-neu-mal', type=str, metavar='C',
                    help='Specification of the NetHack character.')
parser.add_argument("--pipes_basename", default="unix:/tmp/polybeast",
                    help="Basename for the pipes for inter-process communication. "
                    "Has to be of the type unix:/some/path.")
parser.add_argument('--num_servers', default=4, type=int, metavar='N',
                    help='Number of environment servers.')

parser.add_argument('--mock', action="store_true",
                    help='Use mock environment instead of NetHack.')
parser.add_argument('--single_ttyrec', action="store_true",
                    help='Record ttyrec only for actor 0.')
parser.add_argument('--num_seeds', default=0, type=int, metavar='S',
                    help='If larger than 0, samples fixed number of environment seeds '
                         'to be used.')
parser.add_argument('--seedspath', default="", type=str,
                    help="Path to json file with seeds.")

# Training settings.
parser.add_argument('--savedir', default='~/nethackruns',
                    help='Root dir where experiment data will be saved.')

# Task-Specific settings.
parser.add_argument('--reward_win', default=1.0, type=float,
                    help='Reward for winning (finding the staircase).')
parser.add_argument('--reward_lose', default=-1.0, type=float,
                    help='Reward for losing (dying before finding the staircase).')

parser.add_argument('--penalty_step', default=-0.0001, type=float,
                    help='Penalty per step in the episode.')
parser.add_argument('--penalty_time', default=-0.0001, type=float,
                    help='Penalty per time step in the episode.')
parser.add_argument('--fn_penalty_step', default="constant", type=str,
                    help='Function to accumulate penalty.')
parser.add_argument('--max_num_steps', default=1000, type=int,
                    help='Maximum number of steps in the game.')
parser.add_argument('--state_counter', default="none", choices=['none', 'coordinates'],
                    help='Method for counting state visits. Default none. '
                         'Coordinates concatenates dungeon level with player x,y.')
# yapf: enable

logging.basicConfig(
    format=(
        "[%(levelname)s:%(process)d %(module)s:%(lineno)d %(asctime)s] "
        "%(message)s"
    ),
    level=0,
)


def create_folders(flags):
    # Creates some of the folders that would be created by the filewriter.
    logdir = os.path.join(flags.savedir, "archives")
    if not os.path.exists(logdir):
        logging.info("Creating archive directory: %s" % logdir)
        os.makedirs(logdir, exist_ok=True)
    else:
        logging.info("Found archive directory: %s" % logdir)


def serve(flags, server_address, env_id):
    env = lambda: create_env(flags, env_id)
    server = libtorchbeast.Server(env, server_address=server_address)
    server.run()


def main(flags):
    if flags.num_seeds > 0:
        raise NotImplementedError("num_seeds > 0 not currently implemented.")

    create_folders(flags)

    if not flags.pipes_basename.startswith("unix:"):
        raise Exception(
            "--pipes_basename has to be of the form unix:/some/path."
        )

    processes = []
    for i in range(flags.num_servers):
        p = mp.Process(
            target=serve,
            args=(flags, f"{flags.pipes_basename}.{i}", i),
            daemon=True,
        )
        p.start()
        processes.append(p)

    try:
        # We are only here to listen to the interrupt.
        while True:
            time.sleep(10)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    flags = parser.parse_args()
    main(flags)



# Copyright (c) Facebook, Inc. and its affiliates.

from collections import defaultdict
import gym
import numpy as np
import queue
import threading


class CounterWrapper(gym.Wrapper):
    def __init__(self, env, state_counter="none"):
        # intialize state counter
        self.state_counter = state_counter
        if self.state_counter != "none":
            self.state_count_dict = defaultdict(int)
        # this super() goes to the parent of the particular task, not to object
        super().__init__(env)

    def step(self, action):
        # add state counting to step function if desired
        step_return = self.env.step(action)
        if self.state_counter == "none":
            # do nothing
            return step_return

        obs, reward, done, info = step_return

        if self.state_counter == "ones":
            # treat every state as unique
            state_visits = 1
        elif self.state_counter == "coordinates":
            # use the location of the agent in the dungeon to accumulate visits
            features = obs["blstats"]
            x = features[0]
            y = features[1]
            d = features[12]
            coord = (d, x, y)
            self.state_count_dict[coord] += 1
            state_visits = self.state_count_dict[coord]
        else:
            raise NotImplementedError("state_counter=%s" % self.state_counter)

        obs.update(state_visits=np.array([state_visits]))

        if done:
            self.state_count_dict.clear()

        return step_return

    def reset(self, wizkit_items=None):
        # reset state counter when env resets
        obs = self.env.reset(wizkit_items=wizkit_items)
        if self.state_counter != "none":
            self.state_count_dict.clear()
            # current state counts as one visit
            obs.update(state_visits=np.array([1]))
        return obs


class CropWrapper(gym.Wrapper):
    def __init__(self, env, h=9, w=9, pad=0, keys=("tty_chars", "tty_colors")):
        super().__init__(env)
        self.env = env
        self.h = h
        self.w = w
        self.pad = pad
        self.keys = keys
        assert self.h % 2 == 1
        assert self.w % 2 == 1
        self.last_observation = None
        self._actions = self.env._actions

    def render(self, mode="human", crop=True):
        self.env.render()
        obs = self.last_observation
        tty_chars_crop = obs["tty_chars_crop"]
        tty_colors_crop = obs["tty_colors_crop"]
        rendering = self.env.get_tty_rendering(
            tty_chars_crop, tty_colors_crop, print_guides=True
        )
        print(rendering)

    def step(self, action):
        next_state, reward, done, info = self.env.step(action)

        dh = self.h // 2
        dw = self.w // 2

        (y, x) = next_state["tty_cursor"]
        x += dw
        y += dh

        for key in self.keys:
            obs = next_state[key]
            obs = np.pad(
                obs,
                pad_width=(dw, dh),
                mode="constant",
                constant_values=self.pad,
            )
            next_state[key + "_crop"] = obs[
                y - dh : y + dh + 1, x - dw : x + dw + 1
            ]

        self.last_observation = next_state

        return next_state, reward, done, info

    def reset(self, wizkit_items=None):
        obs = self.env.reset(wizkit_items=wizkit_items)
        obs["tty_chars_crop"] = np.zeros((self.h, self.w), dtype=np.uint8)
        obs["tty_colors_crop"] = np.zeros((self.h, self.w), dtype=np.int8)
        self.last_observation = obs
        return obs


class PrevWrapper(gym.Wrapper):
    def __init__(self, env):
        super().__init__(env)
        self.env = env
        self.last_observation = None
        self._actions = self.env._actions

    def step(self, action):
        next_state, reward, done, info = self.env.step(action)
        next_state["prev_reward"] = np.array([reward], dtype=np.float32)
        next_state["prev_action"] = np.array([action], dtype=np.uint8)

        self.last_observation = next_state

        return next_state, reward, done, info

    def reset(self, wizkit_items=None):
        obs = self.env.reset(wizkit_items=wizkit_items)
        obs["prev_reward"] = np.zeros(1, dtype=np.float32)
        obs["prev_action"] = np.zeros(1, dtype=np.uint8)
        self.last_observation = obs
        return obs


def target(resetqueue, readyqueue):
    while True:
        env = resetqueue.get()
        if env is None:
            return
        obs = env.reset()
        readyqueue.put((obs, env))


class CachedEnvWrapper(gym.Env):
    def __init__(self, envs, num_threads=2):
        self._envs = envs

        # This could alternatively also use concurrent.futures. I hesitate to do
        # that as futures.wait would have me deal with sets all the time where they
        # are really not necessary.
        self._resetqueue = queue.SimpleQueue()
        self._readyqueue = queue.SimpleQueue()

        self._threads = [
            threading.Thread(
                target=target, args=(self._resetqueue, self._readyqueue)
            )
            for _ in range(num_threads)
        ]
        for t in self._threads:
            t.start()

        for env in envs[1:]:
            self._resetqueue.put(env)
        self._env = envs[0]

    def reset(self):
        self._resetqueue.put(self._env)
        obs, self._env = self._readyqueue.get()
        return obs

    def step(self, action):
        return self._env.step(action)

    def close(self):
        for _ in self._threads:
            self._resetqueue.put(None)

        for t in self._threads:
            t.join()

        for env in self._envs:
            env.close()

    def seed(self, seed=None):
        self._env.seed(seed)

    def unwrapped(self):
        return self._env

    def __str__(self):
        return "<CachedEnvWrapper envs=%s>" % [str(env) for env in self._envs]

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()
        return False  # Propagate exception.



# Copyright (c) Facebook, Inc. and its affiliates.

from minihack import MiniHack
from minihack.envs import register
from nle import nethack


MOVE_ACTIONS = tuple(nethack.CompassDirection)


class MiniHackNavigation(MiniHack):
    """The base class for MiniHack Navigation tasks.

    Navigation tasks have the following characteristics:

    - Restricted action space: By default, the agent can only move towards
      eight compass directions.
    - Yes/No questions, as well as menu-selection actions are disabled by
      default.
    - The character is set to chaotic human male rogue.
    - Auto-pick is enabled by default.
    - Maximum episode limit defaults to 100 (can be overriden via the
      `max_episode_steps` argument)
    - The default goal is to reach the stair down. This can be changed using
      a reward manager.
    """

    def __init__(self, *args, des_file: str = None, **kwargs):
        # Actions space - move only by default
        kwargs["actions"] = kwargs.pop("actions", MOVE_ACTIONS)
        # Disallowing one-letter menu questions
        kwargs["allow_all_yn_questions"] = kwargs.pop(
            "allow_all_yn_questions", False
        )
        # Perform known steps
        kwargs["allow_all_modes"] = kwargs.pop("allow_all_modes", False)
        # Play with Rogue character by default
        kwargs["character"] = kwargs.pop("character", "rog-hum-cha-mal")
        # Default episode limit
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 100)

        super().__init__(*args, des_file=des_file, **kwargs)


register(
    id="MiniHack-Navigation-Custom-v0",
    entry_point="minihack.navigation:MiniHackNavigation",
)



# Copyright (c) Facebook, Inc. and its affiliates.
from minihack import MiniHackSkill, LevelGenerator, RewardManager
from minihack.envs import register


class MiniHackEat(MiniHackSkill):
    """Environment for "eat" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("apple", "%")
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_eat_event("apple")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackEatFixed(MiniHackSkill):
    """Environment for "eat" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("apple", "%", place=(0, 0))
        lvl_gen.set_start_pos((2, 2))
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_eat_event("apple")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackEatDistr(MiniHackSkill):
    """Environment for "eat" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("apple", "%")
        lvl_gen.add_monster()
        lvl_gen.add_object()
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_eat_event("apple")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackWield(MiniHackSkill):
    """Environment for "wield" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("dagger", ")")
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_wield_event("dagger")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackWieldFixed(MiniHackSkill):
    """Environment for "wield" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("dagger", ")", place=(0, 0))
        lvl_gen.set_start_pos((2, 2))
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_wield_event("dagger")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackWieldDistr(MiniHackSkill):
    """Environment for "wield" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("dagger", ")")
        lvl_gen.add_monster()
        lvl_gen.add_object()
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_wield_event("dagger")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackWear(MiniHackSkill):
    """Environment for "wear" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("robe", "[")
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_wear_event("robe")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackWearFixed(MiniHackSkill):
    """Environment for "wear" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("robe", "[", place=(0, 0))
        lvl_gen.set_start_pos((2, 2))
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_wear_event("robe")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackWearDistr(MiniHackSkill):
    """Environment for "wear" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("robe", "[")
        lvl_gen.add_monster()
        lvl_gen.add_object()
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_wear_event("robe")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackPutOn(MiniHackSkill):
    """Environment for "put on" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("amulet of life saving", '"')
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_amulet_event()

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackPutOnFixed(MiniHackSkill):
    """Environment for "put on" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("amulet of life saving", '"', place=(0, 0))
        lvl_gen.set_start_pos((2, 2))
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_amulet_event()

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackPutOnDistr(MiniHackSkill):
    """Environment for "put on" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("amulet of life saving", '"')
        lvl_gen.add_monster()
        lvl_gen.add_object()
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_amulet_event()

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackZap(MiniHackSkill):
    """Environment for "zap" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("enlightenment", "/")
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_message_event(["The feeling subsides."])

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackZapFixed(MiniHackSkill):
    """Environment for "zap" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("enlightenment", "/", place=(0, 0))
        lvl_gen.set_start_pos((2, 2))
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_message_event(["The feeling subsides."])

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackZapDistr(MiniHackSkill):
    """Environment for "zap" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("enlightenment", "/")
        lvl_gen.add_monster()
        lvl_gen.add_object()
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_message_event(["The feeling subsides."])

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackRead(MiniHackSkill):
    """Environment for "read" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("blank paper", "?")
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_message_event(["This scroll seems to be blank."])

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackReadFixed(MiniHackSkill):
    """Environment for "read" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("blank paper", "?", place=(0, 0))
        lvl_gen.set_start_pos((2, 2))
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_message_event(["This scroll seems to be blank."])

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackReadDistr(MiniHackSkill):
    """Environment for "read" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_object("blank paper", "?")
        lvl_gen.add_monster()
        lvl_gen.add_object()
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_message_event(["This scroll seems to be blank."])

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackPray(MiniHackSkill):
    """Environment for "pray" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_altar("random", "neutral", "altar")
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_positional_event("altar", "pray")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackPrayFixed(MiniHackSkill):
    """Environment for "pray" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_altar((0, 0), "neutral", "altar")
        lvl_gen.set_start_pos((2, 2))
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_positional_event("altar", "pray")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackPrayDistr(MiniHackSkill):
    """Environment for "pray" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_altar("random", "neutral", "altar")
        lvl_gen.add_monster()
        lvl_gen.add_object()
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_positional_event("altar", "pray")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackSink(MiniHackSkill):
    """Environment for "sink" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_sink()
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_positional_event("sink", "quaff")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackSinkFixed(MiniHackSkill):
    """Environment for "sink" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_sink(place=(0, 0))
        lvl_gen.set_start_pos((2, 2))
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_positional_event("sink", "quaff")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackSinkDistr(MiniHackSkill):
    """Environment for "sink" task."""

    def __init__(self, *args, **kwargs):
        lvl_gen = LevelGenerator(w=5, h=5, lit=True)
        lvl_gen.add_sink()
        lvl_gen.add_monster()
        lvl_gen.add_object()
        des_file = lvl_gen.get_des()

        reward_manager = RewardManager()
        reward_manager.add_positional_event("sink", "quaff")

        super().__init__(
            *args, des_file=des_file, reward_manager=reward_manager, **kwargs
        )


class MiniHackClosedDoor(MiniHackSkill):
    """Environment for "open" task."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, des_file="closed_door.des", **kwargs)


class MiniHackLockedDoor(MiniHackSkill):
    """Environment for "kick" task."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, des_file="locked_door.des", **kwargs)


class MiniHackLockedDoorFixed(MiniHackSkill):
    """Environment for "kick" task."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, des_file="locked_door_fixed.des", **kwargs)


# Tasks (w/o doors)
register(
    id="MiniHack-Eat-v0",
    entry_point="minihack.envs.skills_simple:MiniHackEat",
)
register(
    id="MiniHack-Pray-v0",
    entry_point="minihack.envs.skills_simple:MiniHackPray",
)
register(
    id="MiniHack-Sink-v0",
    entry_point="minihack.envs.skills_simple:MiniHackSink",
)
register(
    id="MiniHack-Wield-v0",
    entry_point="minihack.envs.skills_simple:MiniHackWield",
)
register(
    id="MiniHack-Wear-v0",
    entry_point="minihack.envs.skills_simple:MiniHackWear",
)
register(
    id="MiniHack-PutOn-v0",
    entry_point="minihack.envs.skills_simple:MiniHackPutOn",
)
register(
    id="MiniHack-Zap-v0",
    entry_point="minihack.envs.skills_simple:MiniHackZap",
)
register(
    id="MiniHack-Read-v0",
    entry_point="minihack.envs.skills_simple:MiniHackRead",
)

# Fixed version of tasks
register(
    id="MiniHack-Eat-Fixed-v0",
    entry_point="minihack.envs.skills_simple:MiniHackEatFixed",
)
register(
    id="MiniHack-Pray-Fixed-v0",
    entry_point="minihack.envs.skills_simple:MiniHackPrayFixed",
)
register(
    id="MiniHack-Sink-Fixed-v0",
    entry_point="minihack.envs.skills_simple:MiniHackSinkFixed",
)
register(
    id="MiniHack-Wield-Fixed-v0",
    entry_point="minihack.envs.skills_simple:MiniHackWieldFixed",
)
register(
    id="MiniHack-Wear-Fixed-v0",
    entry_point="minihack.envs.skills_simple:MiniHackWearFixed",
)
register(
    id="MiniHack-PutOn-Fixed-v0",
    entry_point="minihack.envs.skills_simple:MiniHackPutOnFixed",
)
register(
    id="MiniHack-Zap-Fixed-v0",
    entry_point="minihack.envs.skills_simple:MiniHackZapFixed",
)
register(
    id="MiniHack-Read-Fixed-v0",
    entry_point="minihack.envs.skills_simple:MiniHackReadFixed",
)

# Task versions with random distractions
register(
    id="MiniHack-Eat-Distr-v0",
    entry_point="minihack.envs.skills_simple:MiniHackEatDistr",
)
register(
    id="MiniHack-Pray-Distr-v0",
    entry_point="minihack.envs.skills_simple:MiniHackPrayDistr",
)
register(
    id="MiniHack-Sink-Distr-v0",
    entry_point="minihack.envs.skills_simple:MiniHackSinkDistr",
)
register(
    id="MiniHack-Wield-Distr-v0",
    entry_point="minihack.envs.skills_simple:MiniHackWieldDistr",
)
register(
    id="MiniHack-Wear-Distr-v0",
    entry_point="minihack.envs.skills_simple:MiniHackWearDistr",
)
register(
    id="MiniHack-PutOn-Distr-v0",
    entry_point="minihack.envs.skills_simple:MiniHackPutOnDistr",
)
register(
    id="MiniHack-Zap-Distr-v0",
    entry_point="minihack.envs.skills_simple:MiniHackZapDistr",
)
register(
    id="MiniHack-Read-Distr-v0",
    entry_point="minihack.envs.skills_simple:MiniHackReadDistr",
)

# Tasks involvign doors
register(
    id="MiniHack-ClosedDoor-v0",
    entry_point="minihack.envs.skills_simple:MiniHackClosedDoor",
)
register(
    id="MiniHack-LockedDoor-v0",
    entry_point="minihack.envs.skills_simple:MiniHackLockedDoor",
)
register(
    id="MiniHack-LockedDoor-Fixed-v0",
    entry_point="minihack.envs.skills_simple:MiniHackLockedDoorFixed",
)



# Copyright (c) Facebook, Inc. and its affiliates.

import os
import subprocess
import random
from typing import Tuple

import gym
import numpy as np
import pkg_resources
from nle import _pynethack, nethack
from nle.nethack.nethack import SCREEN_DESCRIPTIONS_SHAPE, OBSERVATION_DESC
from nle.env.base import FULL_ACTIONS, NLE_SPACE_ITEMS
from nle.env.tasks import NetHackStaircase
from minihack.wiki import NetHackWiki
from minihack.tiles import GlyphMapper

PATH_DAT_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "dat")
LIB_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "lib")
PATCH_SCRIPT = os.path.join(
    pkg_resources.resource_filename("minihack", "scripts"),
    "mh_patch_nhdat.sh",
)
MH_FULL_ACTIONS = list(FULL_ACTIONS)
try:
    MH_FULL_ACTIONS.remove(nethack.MiscDirection.UP)
except ValueError:
    pass
MH_FULL_ACTIONS = tuple(MH_FULL_ACTIONS)
HACKDIR = pkg_resources.resource_filename("nle", "nethackdir")

RGB_MAX_VAL = 255
N_TILE_PIXEL = 16

MH_NETHACKOPTIONS = (
    "color",  # Display color for different monsters, objects, etc
    "showexp",  # Display the experience points on the status line
    "nobones",  # Disallow saving and loading bones files
    "nolegacy",  # Not display an introductory message when starting the game
    "nocmdassist",  # No command assistance
    "disclose:+i +a +v +g +c +o",  # End of game prompt replies
    "runmode:teleport",  # Update the map after movement has finished
    "mention_walls",  # Give feedback when walking against a wall
    "nosparkle",  # Not display sparkly effect for resisted magical attacks
    "showscore",  # Shows approximate accumulated score on the bottom line
)
# Autopickup on by default (all items)
# Manually adding "!autopickup" basen of env flag

MINIHACK_SPACE_FUNCS = {
    "glyphs_crop": lambda x, y: gym.spaces.Box(
        low=0,
        high=nethack.MAX_GLYPH,
        shape=(x, y),
        dtype=OBSERVATION_DESC["glyphs"]["dtype"],
    ),
    "chars_crop": lambda x, y: gym.spaces.Box(
        low=0,
        high=255,
        shape=(x, y),
        dtype=OBSERVATION_DESC["chars"]["dtype"],
    ),
    "colors_crop": lambda x, y: gym.spaces.Box(
        low=0,
        high=15,
        shape=(x, y),
        dtype=OBSERVATION_DESC["colors"]["dtype"],
    ),
    "specials_crop": lambda x, y: gym.spaces.Box(
        low=0,
        high=255,
        shape=(x, y),
        dtype=OBSERVATION_DESC["specials"]["dtype"],
    ),
    "tty_chars_crop": lambda x, y: gym.spaces.Box(
        low=0,
        high=255,
        shape=(x, y),
        dtype=OBSERVATION_DESC["tty_chars"]["dtype"],
    ),
    "tty_colors_crop": lambda x, y: gym.spaces.Box(
        low=0,
        high=31,
        shape=(x, y),
        dtype=OBSERVATION_DESC["tty_colors"]["dtype"],
    ),
    "screen_descriptions_crop": lambda x, y: gym.spaces.Box(
        low=0,
        high=127,
        shape=(x, y, _pynethack.nethack.NLE_SCREEN_DESCRIPTION_LENGTH),
        dtype=OBSERVATION_DESC["screen_descriptions"]["dtype"],
    ),
    "pixel_crop": lambda x, y: gym.spaces.Box(
        low=0,
        high=RGB_MAX_VAL,
        shape=(x * N_TILE_PIXEL, y * N_TILE_PIXEL, 3),
        dtype=np.uint8,
    ),
}

MH_DEFAULT_OBS_KEYS = [
    "glyphs",
    "chars",
    "colors",
    "specials",
    "glyphs_crop",
    "chars_crop",
    "colors_crop",
    "specials_crop",
    "blstats",
    "message",
]


class MiniHack(NetHackStaircase):
    """MiniHack base class.

    All MiniHack environments are derived from this class, which itself is
    derived from NLE base class.

    Note that this class itself is not used for creating new environment
    instances. Instead, ``MiniHackNavigation`` and ``MiniHackSkill`` provide a
    more convenient interface for doing this, both of which are directly
    derived from MiniHack for specific types of environments.
    """

    def __init__(
        self,
        *args,
        des_file: str,
        reward_win=1,
        reward_lose=0,
        obs_crop_h=9,
        obs_crop_w=9,
        obs_crop_pad=0,
        reward_manager=None,
        use_wiki=False,
        autopickup=True,
        pet=False,
        observation_keys=MH_DEFAULT_OBS_KEYS,
        seeds=None,
        **kwargs,
    ):
        """Constructs a new MiniHack environment.

        Args:
            des_file (str):
                The description file for the environment.
            reward_win (float):
                The reward received upon successfully completing an episode.
                Defaults to 1.
            reward_lose (float):
                The reward received upon death or aborting. Defaults to 1.
            obs_crop_h (int):
                The height of agent-centred cropped observation. Defaults to 9.
            obs_crop_w (int):
                The width of agent-centred cropped observation. Defaults to 9.
            obs_crop_pad (int):
                The padding for agent-centred cropped observation.
                Defaults to 0.
            reward_manager (RewardManager or None):
                The reward manager that describes the custom reward function of
                the agent. If None, the goal of the agent is to reach the stair
                down. Defaults to None.
            use_wiki (bool):
                Whether to use the NetHack wiki. Defaults to False.
            autopickup (bool):
                Turning autopickup on or off. Defaults to True.
            pet (bool):
                Whether to include the pet. Defaults to False.
            observation_keys (list):
                The keys of observations returned after every timestep by the
                environment as a dictionary. Defaults to
                ``minihack.base.MH_DEFAULT_OBS_KEYS``.
            seeds (list or None):
                A list of random seeds for sampling episodes. If none, the
                entire level distribution is used. Defaults to None.
            penalty_mode (str):
                The name of the mode for calculating the time step penalty.
                Can be ``constant``, ``exp``, ``square``, ``linear``, or
                ``always``. Defaults to ``constant``. Inherited from
                `NetHackScore`.
            penalty_step (float):
                A constant applied to amount of frozen steps. Defaults to
                -0.01. Inherited from `NetHackScore`.
            penalty_time (float):
                A constant applied to amount of frozen steps. Defaults to -0.0.
                Inherited from `NetHackScore`.
            savedir (str or None): path to save ttyrecs (game recordings) into.
                Defaults to None, which doesn't save any data. Otherwise,
                interpreted as a path to a new or existing directory.
                If "" (empty string), NLE choses a unique directory name.
                Inherited from `NLE`.
            character (str):
                Name of character. Defaults to "mon-hum-neu-mal". Interited
                from `NLE`.
            max_episode_steps (int):
                maximum amount of steps allowed before the game is forcefully
                quit.  In such cases, ``info["end_status"]`` ill be equal to
                ``StepStatus.ABORTED``. Defaults to 5000. Inherited from `NLE`.
            actions (list):
                list of actions. If None, the full action space will
                be used, i.e. ``nle.nethack.ACTIONS``. Defaults to None.
                Inherited from `NLE`.
            wizard (bool):
                activate wizard mode. Defaults to False.
            allow_all_yn_questions (bool):
                If set to True, no y/n questions in step() are declined.
                If set to False, only elements of SKIP_EXCEPTIONS are not
                declined. Defaults to False. Inherited from `NLE`.
            allow_all_modes (bool):
                If set to True, do not decline menus, text input or auto
                'MORE'. If set to False, only skip click through 'MORE'
                on death. Inherited from `NLE`.
        """
        # NetHack options
        options: Tuple = MH_NETHACKOPTIONS
        if not autopickup:
            options += ("!autopickup",)
        if not pet:
            options += ("pettype:none",)
        kwargs["options"] = kwargs.pop("options", options)
        # Actions space - move only
        kwargs["actions"] = kwargs.pop("actions", MH_FULL_ACTIONS)

        # Enter Wizard mode - turned off by default
        kwargs["wizard"] = kwargs.pop("wizard", False)
        # Allowing one-letter menu questions
        kwargs["allow_all_yn_questions"] = kwargs.pop(
            "allow_all_yn_questions", True
        )
        # Episode limit
        kwargs["max_episode_steps"] = kwargs.pop("max_episode_steps", 200)
        # Not saving NLE data by detauls
        kwargs["savedir"] = kwargs.pop("savedir", None)
        # Not spawning random monsters
        kwargs["spawn_monsters"] = kwargs.pop("spawn_monsters", False)

        # MiniHack's observation keys are kept separate
        self._minihack_obs_keys = list(observation_keys)
        # Handle RGB pixel observations
        if any("pixel" in key for key in self._minihack_obs_keys):
            self._glyph_mapper = GlyphMapper()
            # Make sure glyphs_crop is there
            if (
                "pixel_crop" in self._minihack_obs_keys
                and "glyphs_crop" not in self._minihack_obs_keys
            ):
                self._minihack_obs_keys.append("glyphs_crop")

        self.reward_manager = reward_manager
        if self.reward_manager is not None:
            self.reward_manager.reset()

        self._level_seeds = seeds

        super().__init__(*args, **kwargs)

        # Patch the nhdat library by compling the given .des file
        self.update(des_file)

        self.obs_crop_h = obs_crop_h
        self.obs_crop_w = obs_crop_w
        self.obs_crop_pad = obs_crop_pad

        assert self.obs_crop_h % 2 == 1
        assert self.obs_crop_w % 2 == 1

        self.reward_win = reward_win
        self.reward_lose = reward_lose

        self._scr_descr_index = self._observation_keys.index(
            "screen_descriptions"
        )
        self.observation_space = gym.spaces.Dict(
            self._get_obs_space_dict(dict(NLE_SPACE_ITEMS))
        )

        self.use_wiki = use_wiki
        if self.use_wiki:
            self.wiki = NetHackWiki()

    def _get_obs_space_dict(self, space_dict):
        obs_space_dict = {}
        for key in self._minihack_obs_keys:
            if key in space_dict.keys():
                obs_space_dict[key] = space_dict[key]
            elif key in MINIHACK_SPACE_FUNCS.keys():
                space_func = MINIHACK_SPACE_FUNCS[key]
                obs_space_dict[key] = space_func(
                    self.obs_crop_h, self.obs_crop_w
                )
            else:
                if "pixel" in self._minihack_obs_keys:
                    d_shape = OBSERVATION_DESC["glyphs"]["shape"]
                    shape = (
                        d_shape[0] * N_TILE_PIXEL,
                        d_shape[1] * N_TILE_PIXEL,
                        3,
                    )
                    obs_space_dict["pixel"] = gym.spaces.Box(
                        low=0,
                        high=RGB_MAX_VAL,
                        shape=shape,
                        dtype=np.uint8,
                    )
                else:
                    raise ValueError(
                        f'Observation key "{key}" is not supported'
                    )

        return obs_space_dict

    def reset(self, *args, **kwargs):
        if self.reward_manager is not None:
            self.reward_manager.reset()
        if self._level_seeds is not None:
            seed = random.choice(self._level_seeds)
            self.seed(seed, seed, reseed=False)
        return super().reset(*args, **kwargs)

    def _reward_fn(self, last_observation, action, observation, end_status):
        """Use reward_manager to collect reward calculated in _is_episode_end,
        or revert to standard sparse reward.
        """
        del action  # Unused
        if self.reward_manager is not None:
            reward = self.reward_manager.collect_reward()
        else:
            if end_status == self.StepStatus.TASK_SUCCESSFUL:
                reward = self.reward_win
            elif end_status == self.StepStatus.RUNNING:
                reward = 0
            else:  # death or aborted
                reward = self.reward_lose
        return reward + self._get_time_penalty(last_observation, observation)

    def step(self, action: int):
        self._previous_obs = tuple(a.copy() for a in self.last_observation)
        self._previous_action = action
        # Within this call, _is_episode_end is called and then _reward_fn,
        # both using self.reward_manager
        return super().step(action)

    def _is_episode_end(self, observation):
        if self.reward_manager is not None:
            # This also calculates reward, to be collected in _reward_fn by
            # collect_reward
            result = self.reward_manager.check_episode_end_call(
                self, self._previous_obs, self._previous_action, observation
            )
            if result:
                return self.StepStatus.TASK_SUCCESSFUL

        # Revert to staircase check (so we always end if we reach it)
        return super()._is_episode_end(observation)

    def update(self, des_file):
        """Update the current environment by replacing its description file."""
        self._patch_nhdat(des_file)

    def _patch_nhdat(self, des_file):
        """Patch the nhdat library. This includes compiling the given
        description file and replacing the new nhdat file in the temporary
        hackdir directory of the environment.
        """
        if not des_file.endswith(".des"):
            fpath = os.path.join(self.env._vardir, "mylevel.des")
            # If the des-file is passed as a string
            with open(fpath, "w") as f:
                f.writelines(des_file)
            des_file = fpath

        # Use the .des file if exists, otherwise search in minihack directory
        des_path = os.path.abspath(des_file)
        if not os.path.exists(des_path):
            des_path = os.path.abspath(os.path.join(PATH_DAT_DIR, des_file))
        if not os.path.exists(des_path):
            print(
                "{} file doesn't exist. Please provide a path to a valid .des \
                    file".format(
                    des_path
                )
            )
        try:
            _ = subprocess.call(
                [
                    PATCH_SCRIPT,
                    self.env._vardir,
                    HACKDIR,
                    LIB_DIR,
                    des_path,
                ]
            )
        except subprocess.CalledProcessError as e:
            raise RuntimeError(f"Couldn't patch the nhdat file.\n{e}")

    def _get_observation(self, observation):
        # Filter out observations that we don't need
        observation = super()._get_observation(observation)
        obs_dict = {}
        for key in self._minihack_obs_keys:
            if "pixel" in key:
                continue
            if key in self._observation_keys:
                obs_dict[key] = observation[key]
            elif key in MINIHACK_SPACE_FUNCS.keys():
                orig_key = key.replace("_crop", "")
                if "tty" in orig_key:
                    loc = observation["tty_cursor"][::-1]
                else:
                    loc = observation["blstats"][:2]
                obs_dict[key] = self._crop_observation(
                    observation[orig_key], loc
                )

        if "pixel" in self._minihack_obs_keys:
            obs_dict["pixel"] = self._glyph_mapper.to_rgb(
                observation["glyphs"]
            )

        if "pixel_crop" in self._minihack_obs_keys:
            obs_dict["pixel_crop"] = self._glyph_mapper.to_rgb(
                obs_dict["glyphs_crop"]
            )

        return obs_dict

    def _crop_observation(self, obs, loc):
        dh = self.obs_crop_h // 2
        dw = self.obs_crop_w // 2

        (x, y) = loc
        x += dw
        y += dh

        obs = np.pad(
            obs,
            pad_width=(dw, dh),
            mode="constant",
            constant_values=self.obs_crop_pad,
        )
        return obs[y - dh : y + dh + 1, x - dw : x + dw + 1]

    def key_in_inventory(self, name):
        """Returns key of the given object in the inventory.

        Args:
            name (str): Name of the object.
        Returns:
            str: the key of the first item in the inventory that includes the
            argument name as a substring. Returns None if not found.
        """
        assert "inv_strs" in self._observation_keys
        assert "inv_letters" in self._observation_keys

        inv_strs_index = self._observation_keys.index("inv_strs")
        inv_letters_index = self._observation_keys.index("inv_letters")

        inv_strs = self.last_observation[inv_strs_index]
        inv_letters = self.last_observation[inv_letters_index]

        for letter, line in zip(inv_letters, inv_strs):
            if np.all(line == 0):
                break
            if name in line.tobytes().decode("utf-8"):
                return letter.tobytes().decode("utf-8")

        return None

    def _index_to_dir_action(self, index):
        """Returns the ASCII code for direction corresponding to given
        index in reshaped vector of adjacent 9 tiles (None for agent's
        position).
        """
        assert 0 <= index < 9
        index_to_dir_dict = {
            0: ord("y"),
            1: ord("k"),
            2: ord("u"),
            3: ord("h"),
            4: None,
            5: ord("l"),
            6: ord("b"),
            7: ord("j"),
            8: ord("n"),
        }
        return index_to_dir_dict[index]

    def get_object_direction(self, name, observation=None):
        """Find the game direction of the (first) object in the neighboring
        nine tiles that contains the given name in its description.

        Args:
            name (str): Name of the object.
            observation (dict): Agent observation.

        Returns:
            int: The index of the direction. None if not found.
        """
        if observation is None:
            observation = self.last_observation

        neighbors = self.get_neighbor_descriptions(observation)
        for i, tile_description in enumerate(neighbors):
            if name in tile_description:
                return self._index_to_dir_action(i)
        return None

    def get_neighbor_descriptions(self, observation=None):
        """Returns the descriptions of nine neighboring grids around the
        agent.
        """
        if observation is None:
            observation = self.last_observation
        blstats = observation[self._blstats_index]
        x, y = blstats[:2]

        neighbors = [
            self.get_screen_description(i, j, observation)
            for j in range(y - 1, y + 2)
            for i in range(x - 1, x + 2)
        ]
        return neighbors

    def get_neighbor_wiki_pages(self, observation=None):
        """Returns the page contents of the neighboring objects from NetHack
        wiki.
        """
        if not self.use_wiki:
            raise NotImplementedError(
                "use_wiki is set to false - initialise your environment with"
                "use_wiki=True to use the wiki"
            )
        neighbors_descriptions = self.get_neighbor_descriptions(observation)
        neighbor_pages = [
            self.wiki.get_page_text(description)
            for description in neighbors_descriptions
        ]
        return neighbor_pages

    def get_screen_description(self, x, y, observation=None):
        """Returns the description of the screen on (x,y) coordinates."""
        if observation is None:
            observation = self.last_observation

        des_arr = observation[self._scr_descr_index][y, x]
        symb_len = np.where(des_arr == 0)[0][0]

        return des_arr[:symb_len].tobytes().decode("utf-8")

    def get_screen_wiki_page(self, x, y, observation=None):
        """Returns the wiki page matching the object on (x,y) coordinates."""
        if not self.use_wiki:
            raise NotImplementedError(
                "use_wiki is set to false - initialise your environment with"
                "use_wiki=True to use the wiki"
            )
        description = self.get_screen_description(x, y, observation)
        return self.wiki.get_page_text(description)

    def screen_contains(self, name, observation=None):
        """Whether an object with the given name is visible on the screen, i.e.
        included in the screen descriptions of the observation dictionary.

        Args:
            name (str): Name of the object or monster.
            observation (dict): Agent observation.

        Returns:
            bool: True if the name is contained on the screen, False otherwise.
        """
        if observation is None:
            observation = self.last_observation

        y, x = SCREEN_DESCRIPTIONS_SHAPE[0:2]
        for j in range(y):
            for i in range(x):
                des_arr = observation[self._scr_descr_index][j, i]
                symb_len = np.where(des_arr == 0)[0][0]
                if name in des_arr[:symb_len].tobytes().decode("utf-8"):
                    return True
        return False



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from torch import nn
from torch.nn import functional as F


class ForwardDynamicsNet(nn.Module):
    def __init__(self, num_actions, hidden_dim, input_dim, output_dim):
        super(ForwardDynamicsNet, self).__init__()
        self.num_actions = num_actions

        total_input_dim = input_dim + self.num_actions
        self.forward_dynamics = nn.Sequential(
            nn.Linear(total_input_dim, hidden_dim),
            nn.ELU(),
            nn.Linear(hidden_dim, output_dim),
        )

    def forward(self, state_embedding, action):
        action_one_hot = F.one_hot(
            action, num_classes=self.num_actions
        ).float()
        inputs = torch.cat((state_embedding, action_one_hot), dim=-1)
        next_state_emb = self.forward_dynamics(inputs)
        return next_state_emb


class InverseDynamicsNet(nn.Module):
    def __init__(self, num_actions, hidden_dim, input_dim1, input_dim2):
        super(InverseDynamicsNet, self).__init__()
        self.num_actions = num_actions

        total_input_dim = input_dim1 + input_dim2  # concat the inputs
        self.inverse_dynamics = nn.Sequential(
            nn.Linear(total_input_dim, hidden_dim),
            nn.ELU(),
            nn.Linear(hidden_dim, self.num_actions),
        )

    def forward(self, state_embedding, next_state_embedding):
        inputs = torch.cat((state_embedding, next_state_embedding), dim=-1)
        action_logits = self.inverse_dynamics(inputs)
        return action_logits



# Copyright (c) Facebook, Inc. and its affiliates.

import gym
from gym.envs import registration


def register(id, **kwargs):
    if gym.__version__ >= "0.21":
        # Starting with version 0.21, gym wraps everything by the
        # OrderEnforcing wrapper by default (which isn't in gym.wrappers).
        # This breaks our seed() calls and some other code. Disable.
        kwargs["order_enforce"] = False

    registration.register(id, **kwargs)



# Copyright (c) Facebook, Inc. and its affiliates.

from minihack import MiniHackSkill
from nle.nethack.actions import MiscDirection
import PIL.Image
import PIL.ImageChops
import math


def get_image(pixel_obs, full_screen=False, resize=None, relative=False):
    image = PIL.Image.fromarray(pixel_obs)
    if not full_screen:
        bg = PIL.Image.new(image.mode, image.size, image.getpixel((0, 0)))
        diff = PIL.ImageChops.difference(image, bg)
        diff = PIL.ImageChops.add(diff, diff, 2.0, -100)
        bbox = diff.getbbox()
        if bbox:
            image = image.crop(bbox)

    if resize is not None:
        x_resize, y_resize = resize
        if relative:
            image = image.resize(
                (int(image.width * x_resize), int(image.height * y_resize))
            )
        else:
            image = image.resize((x_resize, y_resize))

    return image


def get_des_file_rendering(
    des_file,
    n_images=1,
    full_screen=False,
    full_obs=True,
    resize=None,
    relative=True,
    wizard=False,
):
    class MHCustom(MiniHackSkill):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, des_file=des_file, **kwargs)

    def get_pixel_obs(env, full_obs=True):
        obs = env.reset()
        if full_obs and wizard:
            for c in (
                "#wizintrinsic\rt\r\r#wizmap\r#wizwish\r"
                + "a potion of object detection\r"
            ):
                obs, sds = env.env.step(ord(c))
            msg = (
                obs[env._original_observation_keys.index("message")]
                .tobytes()
                .decode("utf-8")
            )

            for c in f"q{msg[0]}":
                obs, sds = env.env.step(ord(c))

            obs, _, _, _ = env.step(env._actions.index(MiscDirection.WAIT))
        return obs

    env = MHCustom(
        savedir=None,
        archivefile=None,
        observation_keys=("pixel",),
        wizard=wizard,
    )
    if n_images == 1:
        obs = get_pixel_obs(env, full_obs=full_obs)
        image = get_image(
            obs["pixel"],
            full_screen=full_screen,
            resize=resize,
            relative=relative,
        )
        return image
    else:
        images = []
        for _i in range(n_images):
            obs = get_pixel_obs(env, full_obs=full_obs)
            images.append(
                get_image(
                    obs["pixel"],
                    resize=resize,
                    relative=relative,
                    full_screen=full_screen,
                )
            )
        width = images[0].width
        height = images[0].height
        result = PIL.Image.new(
            "RGB",
            ((width * 3) + 10, ((height + 5) * math.ceil(n_images / 3)) - 5),
            color=(200, 200, 200),
        )
        for j in range(math.ceil(n_images / 3)):
            for i in range(3):
                try:
                    result.paste(
                        images[(j * 3) + i],
                        ((width + 5) * i, (height + 5) * j),
                    )
                except IndexError:
                    return result
        return result



#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
import multiprocessing as mp

import gym
import pytest

import minihack  # noqa: F401

START_METHODS = [
    m
    for m in ("fork", "spawn")
    if m in mp.get_all_start_methods()
]


def new_env_one_step():
    env = gym.make("MiniHack-Room-15x15-v0")
    env.reset()
    obs, reward, done, _ = env.step(0)
    return done


@pytest.mark.parametrize(
    "ctx", [mp.get_context(m) for m in START_METHODS], ids=START_METHODS
)
class TestEnvSubprocess:
    def test_env_in_subprocess(self, ctx):
        p = ctx.Process(target=new_env_one_step)
        p.start()
        p.join()
        assert p.exitcode == 0

    def test_env_before_and_in_subprocess(self, ctx):
        new_env_one_step()
        p = ctx.Process(target=new_env_one_step)
        p.start()
        p.join()
        assert p.exitcode == 0



# Copyright (c) Facebook, Inc. and its affiliates.
import numpy as np
import os

PATH_DAT_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "dat")

MAZE_FLAGS = (
    "noteleport",
    "hardfloor",
    "nommap",
    "arboreal",
    "shortsighted",
    "mazelevel",
    "premapped",
    "shroud",
    "graveyard",
    "icedpools",
    "solidify",
    "corrmaze",
    "inaccessibles",
)

MAP_CHARS = [
    " ",  # solid wall
    "#",  # corridor
    ".",  # room floor
    "-",  # horizontal wall
    "|",  # vertical wall
    "+",  # door
    "A",  # air
    "B",  # crosswall
    "C",  # cloud
    "S",  # secret door
    "H",  # secret corridor
    "{",  # fountain
    "\\",  # throne
    "K",  # soml
    "}",  # moat
    "P",  # pool of water
    "L",  # lava pool
    "I",  # ice
    "W",  # water
    "T",  # tree
    "F",  # iron bars
]

TRAP_NAMES = [
    "anti magic",
    "arrow",
    "bear",
    "board",
    "dart",
    "falling rock",
    "fire",
    "hole",
    "land mine",
    "level teleport",
    "magic portal",
    "magic",
    "pit",
    "polymorph",
    "rolling boulder",
    "rust",
    "sleep gas",
    "spiked pit",
    "statue",
    "teleport",
    "trap door",
    "web",
]


class LevelGenerator:
    """
    LevelGenerator provides a convenient Python interface for quickly writing
    description files for MiniHack. The LevelGenerator class can be used to
    create MAZE-type levels with specified heights and widths, and can then
    fill those levels with objects, monsters and terrain, and specify the start
    point of the level.

    Args:
        map (str or None):
            The description of the map block of the environment. If None, the
            map will have a rectangle layout with the given height and width.
            Defaults to None.
        w (int):
            The width of map. Only used when `map=None`. Defaults to 8.
        h (int):
            The height of map. Only used when `map=None`. Defaults to 8.
        fill (str):
            A character describing the environment feature that fills the
            map. Only used when ``map=None``. Defaults to ".", which
            corresponds to floor.
        lit (bool):
            Whether the layout is lit or not. This affects the observations the
            agent will receive. If an area is not lit, the agent can only see
            directly adjacent grids. Defaults to True.
        flags (tuple):
            Flags of the environment. For the full list, see
            https://nethackwiki.com/wiki/Des-file_format#FLAGS.
            Defaults to ("hardfloor",).
        solidfill (str):
            A character describing the environment feature used for filling
            solid / unspecified parts of the map. Defaults to " ", which
            corresponds to solid wall.
    """

    def __init__(
        self,
        map=None,
        w=8,
        h=8,
        fill=".",
        lit=True,
        flags=("hardfloor",),
        solidfill=" ",
    ):
        assert all(
            f in MAZE_FLAGS for f in flags
        ), f"One of the provided maze flags is incorrect: {flags}"
        flags_str = ",".join(flags)

        self.header = f"""
MAZE: "mylevel", ' '
FLAGS:{flags_str}
INIT_MAP: solidfill,'{solidfill}'
GEOMETRY:center,center
"""
        self.mapify = lambda x: "MAP\n" + x + "ENDMAP\n"
        self.init_map(map, w, h, fill)

        litness = "lit" if lit else "unlit"
        self.footer = f'REGION:(0,0,{self.x},{self.y}),{litness},"ordinary"\n'

        self.stair_up_exist = False

    def init_map(self, map=None, x=8, y=8, fill="."):
        """Initialise the map block of the des-file."""
        if map is None:
            # Creating empty area
            self.x = x
            self.y = y
            self.map = np.array([[fill] * x] * y, dtype=str)
        else:
            lines = [list(line) for line in map.split("\n") if len(line) > 0]
            self.y = len(lines)
            self.x = max(len(line) for line in lines)
            new_lines = [line + [" "] * (self.x - len(line)) for line in lines]
            self.map = np.array(new_lines)

    def get_map_str(self):
        """Returns the map as a string."""
        map_list = [
            "".join(self.map[i]) + "\n" for i in range(self.map.shape[0])
        ]
        return "".join(map_list)

    def get_map_array(self):
        """Returns the map as a two-dimensional numpy array."""
        return self.map

    def get_des(self):
        """Returns the description file.

        Returns:
            str: the description file as a string.
        """
        return self.header + self.mapify(self.get_map_str()) + self.footer

    @staticmethod
    def _validate_place(place):
        """Validate a given place argument."""
        if place is None:
            place = "random"
        elif isinstance(place, tuple):
            place = LevelGenerator._validate_coord(place)
            place = str(place)
        elif isinstance(place, str):
            pass
        else:
            raise ValueError("Invalid place provided.")

        return place

    @staticmethod
    def _validate_coord(coord):
        """Validate a given typle of coordinates."""
        assert (
            isinstance(coord, tuple)
            and len(coord) == 2
            and isinstance(coord[0], int)
            and isinstance(coord[1], int)
        )
        return coord

    def add_object(
        self, name="random", symbol="%", place=None, cursestate=None
    ):
        """Add an object to the map.

        Args:
            name (str):
                The name of the object. Defaults to random.
            symbol (str):
                The symbol of the object. The symbol should correspond to the
                given object name. For example, "%" symbol corresponds to
                comestibles, so the name of the object should also correspond
                to commestibles (e.g. apple). Not used when name is "random".
                Defaults to "%".
            place (None, tuple or str):
                The place of the added object. If None, the location is
                selected randomly. Tuple values are used for providing exact
                (x, y) coordinates. String values are copied to des-file as is.
                Defaults to None.
            cursetstate (str or None):
                The cursed state of the object. Can be "blessed", "uncursed",
                "cursed" or "random". Defaults to None (not used).
        """
        place = self._validate_place(place)
        assert isinstance(symbol, str) and len(symbol) == 1
        assert isinstance(
            name, str
        )  # TODO maybe check object exists in NetHack

        if name != "random":
            name = f'"{name}"'
            self.footer += f"OBJECT:('{symbol}',{name}),{place}"

            if cursestate is not None:
                assert cursestate in [
                    "blessed",
                    "uncursed",
                    "cursed",
                    "random",
                ]
                if cursestate != "random":
                    self.footer += f",{cursestate}"

            self.footer += "\n"

        else:
            self.footer += f"OBJECT:random,{place}\n"

    def add_object_area(
        self, area_name, name="random", symbol="%", cursestate=None
    ):
        """Add an object in an area of the map defined by `area_name` variable.
        See ``add_object`` for more details.
        """

        place = f"rndcoord({area_name})"
        self.add_object(name, symbol, place, cursestate)

    def add_monster(self, name="random", symbol=None, place=None, args=()):
        """Add a monster to the map.

        Args:
            name (str):
                The name of the monster. Defaults to random.
            symbol (str or None):
                The symbol of the monster. The symbol should correspond to the
                family of the specified mosnter. For example, "d" symbol
                corresponds to canine monsters, so the name of the object should
                also correspond to canines  (e.g. jackal). Not used when name is
                "random". Defaults to None.
            place (None, tuple or str):
                The place of the added object. If None, the location is
                selected randomly. Tuple values are used for providing exact
                (x, y) coordinates. String values are copied to des-file as is.
                Defaults to None.
            args (tuple):
                Additional monster arguments, e.g. "hostile" or "peaceful",
                "asleep" or "awake", etc. For more details, see
                https://nethackwiki.com/wiki/Des-file_format#MONSTER.
        """
        place = self._validate_place(place)
        assert (
            symbol == "random"
            or symbol is None
            or (isinstance(symbol, str) and len(symbol) == 1)
        )
        assert isinstance(
            name, str
        )  # TODO maybe check object exists in NetHac

        if name != "random":
            name = f'"{name}"'

        if symbol is not None:
            name = f"('{symbol}',{name})"

        self.footer += f"MONSTER:{name},{place}"

        if len(args) > 0:
            assert any(
                arg in ["hostile", "peaceful", "asleep", "awake"]
                for arg in args
            )
            for arg in args:
                self.footer += f",{arg}"

        self.footer += "\n"

    def add_terrain(self, coord, flag, in_footer=False):
        """Add terrain features to the map.

        Args:
            coord (tuple):
                A tuple with length two representing the (x, y) coordinates.
            flag (str):
                The flag corresponding to the desired terrain feature. Should
                belong to minihack.level_generator.MAP_CHARS. For more details,
                see https://nethackwiki.com/wiki/Des-file_format#Map_characters
            in_footer (bool):
                Whether to define the terrain feature as an additional line
                in the description file (True) or directly modify the map block
                with the given flag (False). Defaults to False.
        """
        coord = self._validate_coord(coord)
        assert flag in MAP_CHARS

        if in_footer:
            self.footer += f"TERRAIN: {str(coord)}, '{flag}'\n"
        else:
            x, y = coord
            self.map[y, x] = flag

    def fill_terrain(
        self,
        type,
        flag,
        x1,
        y1,
        x2,
        y2,
    ):
        """Fill the areas between (x1, y1) and (x2, y2) with the given dungeon
        feature:

        Args:
            type (str):
                The type of filling. "rect" indicates an unfilled rectangle,
                containing just the edges and none of the interior points.
                "fillrect" denotes filled rectangle containing the edges and
                all of the interior points. "line" is used for a straight line
                drawn from one pair of coordinates to the other using
                Bresenham's line algorithm.
            flag (str):
                The flag corresponding to the desired terrain feature. Should
                belong to minihack.level_generator.MAP_CHARS. For more details,
                see https://nethackwiki.com/wiki/Des-file_format#Map_characters
            x1 (int): x coordinate of point 1.
            y1 (int): y coordinate of point 1.
            x2 (int): x coordinate of point 2.
            y2 (int): y coordinate of point 2.
        """
        assert type in ("rect", "fillrect", "line")
        assert flag in MAP_CHARS
        self.footer += f"TERRAIN:{type} ({x1},{y1},{x2},{y2}),'{flag}'\n"

    def set_area_variable(
        self,
        var_name,  # Should start with $ sign
        type,
        x1,
        y1,
        x2,
        y2,
    ):
        """Set a variable representing an area on the map.

        Args:
            var_name (str):
                The name of the variable.
            type (str):
                The type of filling. "rect" indicates an unfilled rectangle,
                containing just the edges and none of the interior points.
                "fillrect" denotes filled rectangle containing the edges and
                all of the interior points. "line" is used for a straight line
                drawn from one pair of coordinates to the other using
                Bresenham's line algorithm.
            x1 (int): x coordinate of point 1.
            y1 (int): y coordinate of point 1.
            x2 (int): x coordinate of point 2.
            y2 (int): y coordinate of point 2.
        """

        assert type in ("rect", "fillrect", "line")
        if var_name[0] != "$":
            var_name = "$" + var_name
        self.footer += f"{var_name} = selection:{type} ({x1},{y1},{x2},{y2})\n"

    def add_goal_pos(self, place=None):
        """Add a goal at the given place. Same as `add_stair_down`."""
        self.add_stair_down(place)

    def add_stair_down(self, place=None):
        """Add a stair down at the given place.

        Args:
            place (None, tuple or str):
                The place of the added object. If None, the location is
                selected randomly. Tuple values are used for providing exact
                (x, y) coordinates. String values are copied to des-file as is.
                Defaults to None.
        """
        place = self._validate_place(place)
        self.footer += f"STAIR:{place},down\n"

    def set_start_pos(self, coord):
        """Set the starting position of the agent.

        Args:
            coord (tuple):
                A tuple with length two representing the (x, y) coordinates.
        """
        self._add_stair_up(coord)

    def set_start_rect(self, p1, p2):
        """Set the starting position of the agent.

        Args:
            coord (tuple):
                A tuple with length two representing the (x, y) coordinates.
        """
        self._add_stair_up_rect(p1, p2)

    def _add_stair_up(self, coord):
        """Add a stair up at a given coordanate."""
        if self.stair_up_exist:
            return
        x, y = self._validate_coord(coord)
        _x, _y = abs(x - 1), abs(y - 1)  # any different coordinate than (x,y)
        self.footer += f"BRANCH:({x},{y},{x},{y}),({_x},{_y},{_x},{_y})\n"
        self.stair_up_exist = True

    def _add_stair_up_rect(self, p1, p2):
        """Add a stair up at a given rectangle."""
        if self.stair_up_exist:
            return
        x1, y1 = self._validate_coord(p1)
        x2, y2 = self._validate_coord(p2)
        self.footer += f"BRANCH:({x1},{y1},{x2},{y2}),({0},{0},{0},{0})\n"
        self.stair_up_exist = True

    def add_door(self, state, place=None):
        """Add a door.

        Args:
            state (str):
                The state of the door. Possible values are "locked", "closed",
                "open", "nodoor", and "random". Defaults to "random".
            place (None, tuple or str):
                The place of the added object. If None, the location is
                selected randomly. Tuple values are used for providing exact
                (x, y) coordinates. String values are copied to des-file as is.
                Defaults to None.
        """
        place = self._validate_place(place)
        assert state in ["nodoor", "locked", "closed", "open", "random"]
        self.footer += f"DOOR:{state},{place}\n"

    def add_altar(self, place=None, align="random", type="random"):
        """Add an altar.

        Args:
            place (None, tuple or str):
                The place of the added object. If None, the location is
                selected randomly. Tuple values are used for providing exact
                (x, y) coordinates. String values are copied to des-file as is.
                Defaults to None.
            align (str):
                The alignment. Possible values are "noalign", "law", "neutral",
                "chaos", "coaligned", "noncoaligned", and "random". Defaults
                to "random".
            type (str):
                The type of the altar. Possible values are "sanctum", "shrine",
                "altar", and "random". Defaults to random.
        """
        place = self._validate_place(place)
        assert align in [
            "noalign",
            "law",
            "neutral",
            "chaos",
            "coaligned",
            "noncoaligned",
            "random",
        ]
        assert type in ["sanctum", "shrine", "altar", "random"]
        self.footer += f"ALTAR:{place},{align},{type}\n"

    def add_sink(self, place=None):
        """Add a sink.

        Args:
            place (None, tuple or str):
                The place of the added object. If None, the location is
                selected randomly. Tuple values are used for providing exact
                (x, y) coordinates. String values are copied to des-file as is.
                Defaults to None.
        """
        place = self._validate_place(place)
        self.footer += f"SINK:{place}\n"

    def add_trap(self, name="teleport", place=None):
        """Add a trap.

        Args:
            name (str):
                The name of the trap. For possible values, see
                `minihack.level_generator.TRAP_NAMES`. Defaults to "teleport".
            place (None, tuple or str):
                The place of the added object. If None, the location is
                selected randomly. Tuple values are used for providing exact
                (x, y) coordinates. String values are copied to des-file as is.
                Defaults to None.
        """
        place = self._validate_place(place)
        assert name in TRAP_NAMES
        self.footer += f'TRAP:"{name}",{place}\n'

    def add_fountain(self, place=None):
        """Add a fountain.

        Args:
            place (None, tuple or str):
                The place of the added object. If None, the location is
                selected randomly. Tuple values are used for providing exact
                (x, y) coordinates. String values are copied to des-file as is.
                Defaults to None.
        """
        place = self._validate_place(place)
        self.footer += f"FOUNTAIN: {place}\n"

    def add_gold(self, amount, place=None):
        """Add gold on the floor.

        Args:
            amount (int):
                The amount of gold.
            place (None, tuple or str):
                The place of the added object. If None, the location is
                selected randomly. Tuple values are used for providing exact
                (x, y) coordinates. String values are copied to des-file as is.
                Defaults to None.
        """
        place = self._validate_place(place)
        assert amount > 0
        self.footer += f"GOLD: {amount},{place}\n"

    def add_boulder(self, place=None):
        """Add gold on the floor.

        Args:
            amount (int):
                The amount of gold.
            place (None, tuple or str):
                The place of the added object. If None, the location is
                selected randomly. Tuple values are used for providing exact
                (x, y) coordinates. String values are copied to des-file as is.
                Defaults to None.
        """
        place = self._validate_place(place)
        self.footer += f'OBJECT: "boulder", {place}\n'

    def wallify(self):
        """Wallify the map. Turns walls completely surrounded by other walls
        into solid stone ' '.
        """
        self.footer += "WALLIFY\n"

    def add_mazewalk(self, coord=None, dir="east"):
        """Creates a random maze, starting from the given coordinate.

        Mazewalk turns map grids with solid stone into floor. From the starting
        position, it checks the mapgrid in the direction given, and if it's
        solid stone, it will move there, and turn that place into floor. Then
        it will choose a random direction, jump over the nearest mapgrid in that
        direction, and check the next mapgrid for solid stone. If there is solid
        stone, mazewalk will move that direction, changing that place and the
        intervening mapgrid to floor. Normally the generated maze will not have
        any loops.

        Pointing mazewalk at that will create a small maze of trees, but unless
        the map (at the place where it's put into the level) is surrounded by
        something else than solid stone, mazewalk will get out of that MAP.
        Substituting floor characters for some of the trees "in the maze" will
        make loops in the maze, which are not otherwise possible. Substituting
        floor characters for some of the trees at the edges of the map
        will make maze entrances and exits at those places.

        For more details see
        https://nethackwiki.com/wiki/Des-file_format#MAZEWALK.

        Args:
             coord (tuple or None):
                 A tuple with length two representing the (x, y) coordinates.
                 If None is passed, the middle point of the map is selected.
                 Defaults to None.
            dir (str):
                The direction of the start of the maze. Possible values are
                "north", "east", "south", and "east". Defaults to "east".
        """
        if coord is not None:
            x, y = self._validate_coord(coord)
        else:
            x, y = self.x // 2, self.y // 2

        self.footer += f"MAZEWALK:({x},{y}),{dir}\n"

    def add_line(self, str):
        """Add a custom string to the buttom of the description file.

        Args:
            str (str):
                The string to be concatenated to the des-file.
        """
        self.footer += str + "\n"



#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
import os
import random
import sys
import tempfile

import gym
import numpy as np
import pytest

import minihack
import nle
from nle import nethack


def get_minihack_env_ids():
    specs = gym.envs.registry.all()
    skip_envs_list = [
        "MiniHack-Navigation-Custom-v0",
        "MiniHack-Skill-Custom-v0",
    ]
    skip_env_substring = [
        "Boxoban",
        "MultiRoom",
        "SimpleCrossingS",
        "LavaCrossingS",
    ]
    return [
        spec.id
        for spec in specs
        if spec.id.startswith("MiniHack")
        and spec.id not in skip_envs_list
        and all(env_sub not in spec.id for env_sub in skip_env_substring)
    ]


def rollout_env(env, max_rollout_len):
    """Produces a rollout and asserts step outputs.

    Returns final reward. Does not assume that the environment has already been
    reset.
    """
    obs = env.reset()
    assert env.observation_space.contains(obs)

    for _ in range(max_rollout_len):
        a = env.action_space.sample()
        obs, reward, done, info = env.step(a)
        assert env.observation_space.contains(obs)
        assert isinstance(reward, float)
        assert isinstance(done, bool)
        assert isinstance(info, dict)
        if done:
            break
    env.close()
    return reward


def term_screen(obs):
    return "\n".join("".join(chr(c) for c in row) for row in obs["tty_chars"])


def compare_rollouts(env0, env1, max_rollout_len):
    """Checks that two active environments return the same rollout.

    Assumes that the environments have already been reset.
    """
    step = 0
    while True:
        a = env0.action_space.sample()
        obs0, reward0, done0, info0 = env0.step(a)
        obs1, reward1, done1, info1 = env1.step(a)
        step += 1

        s0, s1 = term_screen(obs0), term_screen(obs1)
        top_ten_msg = "You made the top ten list!"
        if top_ten_msg in s0:
            assert top_ten_msg in s1
        else:
            np.testing.assert_equal(obs0, obs1)
        assert reward0 == reward1
        assert done0 == done1

        assert info0 == info1

        if done0 or step >= max_rollout_len:
            return


@pytest.mark.parametrize("env_name", get_minihack_env_ids())
@pytest.mark.parametrize("wizard", [False, True])
class TestGymEnv:
    @pytest.fixture(autouse=True)  # will be applied to all tests in class
    def make_cwd_tmp(self, tmpdir):
        """Makes cwd point to the test's tmpdir."""
        with tmpdir.as_cwd():
            yield

    def test_init(self, env_name, wizard):
        """Tests default initialization given standard env specs."""
        env = gym.make(env_name, wizard=wizard)
        del env

    def test_reset(self, env_name, wizard):
        """Tests default initialization given standard env specs."""
        env = gym.make(env_name, wizard=wizard)
        obs = env.reset()
        assert env.observation_space.contains(obs)

    def test_chars_colors_specials(self, env_name, wizard):
        env = gym.make(
            env_name,
            observation_keys=("chars", "colors", "specials", "blstats"),
        )
        obs = env.reset()

        assert "specials" in obs
        x, y = obs["blstats"][:2]

        # That's where you're @.
        assert obs["chars"][y, x] == ord("@")

        # You're bright (4th bit, 8) white (7), too.
        assert obs["colors"][y, x] == 8 ^ 7

    def test_default_wizard_mode(self, env_name, wizard):
        if wizard:
            env = gym.make(env_name, wizard=wizard)
            assert "playmode:debug" in env.env._options
        else:
            # do not send a parameter to test a default
            env = gym.make(env_name)
            assert "playmode:debug" not in env.env._options


@pytest.mark.parametrize(
    "env_name",
    [
        e
        for e in get_minihack_env_ids()
        if "MazeWalk" in e or "WoD" in e or "Room" in e or "Eat" in e
    ],
)
@pytest.mark.parametrize("rollout_len", [500])
class TestGymEnvRollout:
    @pytest.fixture(autouse=True)  # will be applied to all tests in class
    def make_cwd_tmp(self, tmpdir):
        """Makes cwd point to the test's tmpdir."""
        with tmpdir.as_cwd():
            yield

    def test_rollout(self, env_name, rollout_len):
        """Tests rollout_len steps (or until termination) of random policy."""
        with tempfile.TemporaryDirectory() as savedir:
            env = gym.make(env_name, savedir=savedir)
            rollout_env(env, rollout_len)
            env.close()

            assert os.path.exists(
                os.path.join(savedir, "nle.%i.0.ttyrec.bz2" % os.getpid())
            )

    def test_rollout_no_archive(self, env_name, rollout_len):
        """Tests rollout_len steps (or until termination) of random policy."""
        env = gym.make(env_name, savedir=None)
        assert env.savedir is None
        assert env._stats_file is None
        assert env._stats_logger is None
        rollout_env(env, rollout_len)

    def test_seed_interface_output(self, env_name, rollout_len):
        """Tests whether env.seed output can be reused correctly."""

        env0 = gym.make(env_name)
        env1 = gym.make(env_name)

        seed_list0 = env0.seed()
        env0.reset()

        assert env0.get_seeds() == seed_list0

        seed_list1 = env1.seed(*seed_list0)
        assert seed_list0 == seed_list1

    def test_seed_rollout_seeded(self, env_name, rollout_len):
        """Tests that two seeded envs return same step data."""

        observation_keys = (
            "tty_chars",
            "glyphs",
            "chars",
            "specials",
            "colors",
        )
        env0 = gym.make(env_name, observation_keys=observation_keys)
        env1 = gym.make(env_name, observation_keys=observation_keys)

        env0.seed(123456, 789012)
        obs0 = env0.reset()
        seeds0 = env0.get_seeds()

        assert seeds0 == (123456, 789012, False)

        env1.seed(*seeds0)
        obs1 = env1.reset()
        seeds1 = env1.get_seeds()

        assert seeds0 == seeds1

        np.testing.assert_equal(obs0, obs1)
        compare_rollouts(env0, env1, rollout_len)

    def test_seed_rollout_seeded_int(self, env_name, rollout_len):
        """Tests that two seeded envs return same step data."""

        observation_keys = (
            "tty_chars",
            "glyphs",
            "chars",
            "specials",
            "colors",
        )
        env0 = gym.make(env_name, observation_keys=observation_keys)
        env1 = gym.make(env_name, observation_keys=observation_keys)

        initial_seeds = (
            random.randrange(sys.maxsize),
            random.randrange(sys.maxsize),
            False,
        )
        env0.seed(*initial_seeds)
        obs0 = env0.reset()
        seeds0 = env0.get_seeds()

        env1.seed(*seeds0)
        obs1 = env1.reset()
        seeds1 = env1.get_seeds()

        assert seeds0 == seeds1 == initial_seeds

        np.testing.assert_equal(obs0, obs1)
        compare_rollouts(env0, env1, rollout_len)

    def test_render_ansi(self, env_name, rollout_len):
        env = gym.make(env_name)
        env.reset()
        for _ in range(rollout_len):
            action = env.action_space.sample()
            _, _, done, _ = env.step(action)
            if done:
                env.reset()
            output = env.render(mode="ansi")
            assert isinstance(output, str)
            assert len(output.replace("\n", "")) == np.prod(
                nle.env.DUNGEON_SHAPE
            )


class TestRoomReward:
    """Tests a few game dynamics."""

    @pytest.fixture(autouse=True)  # will be applied to all tests in class
    def make_cwd_tmp(self, tmpdir):
        """Makes cwd point to the test's tmpdir."""
        with tmpdir.as_cwd():
            yield

    @pytest.fixture
    def env(self):
        e = gym.make("MiniHack-Room-5x5-v0")
        try:
            yield e
        finally:
            e.close()

    def test_reward(self, env):
        _ = env.reset()

        for _ in range(4):
            _, reward, done, _ = env.step(env._actions.index(ord("j")))
            assert reward == 0.0
            assert not done

        for _ in range(3):
            _, reward, done, _ = env.step(env._actions.index(ord("l")))
            assert reward == 0.0
            assert not done

        # Hack to quit.
        _, reward, done, _ = env.step(env._actions.index(ord("l")))

        assert done
        assert reward == 1.0



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from torch import nn
import torch
from nle import nethack as nh

from typing import NamedTuple, Union
from collections import namedtuple
from minihack.agent.common.util.id_pairs import id_pairs_table
import logging

Ratio = Union[int, bool]


class Targets(NamedTuple):
    """Class for configuring whch ids you want to embed into the single
    GlyphEmbedding, and in what ratios. The ratio is only relevant if
    do_linear_layer is false, and the embedding is pure concatenation.
    """

    glyphs: Ratio = 0
    groups: Ratio = 0
    subgroup_ids: Ratio = 0
    colors: Ratio = 0
    chars: Ratio = 0
    specials: Ratio = 0
    do_linear_layer: bool = True

    def count_matrices(self):
        """Count of matrices required"""
        return sum(self) - int(self.do_linear_layer)


GLYPH_TYPE_STRATEGIES = {
    "full": Targets(glyphs=True),
    "group_id": Targets(groups=True, subgroup_ids=True),
    "color_char": Targets(colors=True, chars=True, specials=True),
    "all": Targets(
        groups=True, subgroup_ids=True, colors=True, chars=True, specials=True
    ),
    "all_cat": Targets(
        groups=1,
        subgroup_ids=3,
        colors=1,
        chars=2,
        specials=1,
        do_linear_layer=False,
    ),
}


class GlyphEmbedding(nn.Module):
    """Take the glyph information and return an embedding vector."""

    def __init__(
        self, glyph_type, dimension, device=None, use_index_select=None
    ):
        super(GlyphEmbedding, self).__init__()
        logging.debug("Emdedding on device: %s ", device)
        self.glyph_type = glyph_type
        self.use_index_select = use_index_select
        self.device = device
        self.dim = dimension

        if glyph_type not in GLYPH_TYPE_STRATEGIES:
            raise RuntimeError("unexpected glyph_type=%s" % self.glyph_type)
        strategy = GLYPH_TYPE_STRATEGIES[glyph_type]
        self.strategy = strategy

        self._unit_dim = dimension // strategy.count_matrices()
        self._remainder_dim = (
            self.dim - self._unit_dim * strategy.count_matrices()
        )

        if self.requires_id_pairs_table:
            self.register_buffer(
                "_id_pairs_table", torch.from_numpy(id_pairs_table())
            )
        else:
            self._id_pairs_table = None

        # Build our custom embedding matrices
        embed = {}
        if strategy.glyphs:
            embed["glyphs"] = nn.Embedding(
                nh.MAX_GLYPH, self._dim(strategy.glyphs)
            )
        if strategy.colors:
            embed["colors"] = nn.Embedding(16, self._dim(strategy.colors))
        if strategy.chars:
            embed["chars"] = nn.Embedding(256, self._dim(strategy.chars))
        if strategy.specials:
            embed["specials"] = nn.Embedding(256, self._dim(strategy.specials))
        if strategy.groups:
            num_groups = self.id_pairs_table.select(1, 1).max().item() + 1
            embed["groups"] = nn.Embedding(
                num_groups, self._dim(strategy.groups)
            )
        if strategy.subgroup_ids:
            num_subgroup_ids = (
                self.id_pairs_table.select(1, 0).max().item() + 1
            )
            embed["subgroup_ids"] = nn.Embedding(
                num_subgroup_ids, self._dim(strategy.subgroup_ids)
            )

        self.embeddings = nn.ModuleDict(embed)
        self.targets = list(embed.keys())
        self.GlyphTuple = namedtuple("GlyphTuple", self.targets)

        if strategy.do_linear_layer and strategy.count_matrices() > 1:
            self.linear = nn.Linear(
                strategy.count_matrices() * self.dim, self.dim
            )

        if device is not None:
            self.to(device)

    def _dim(self, units):
        """Decide the embedding size for a single matrix.  If using a linear layer
        at the end this is always the embedding dimension, otherwise it is a
        fraction of the embedding dim"""
        if self.strategy.do_linear_layer:
            return self.dim
        else:
            dim = units * self._unit_dim + self._remainder_dim
            self._remainder_dim = 0
            return dim

    @property
    def requires_id_pairs_table(self):
        return self.strategy.groups or self.strategy.subgroup_ids

    @property
    def id_pairs_table(self):
        return self._id_pairs_table

    def prepare_input(self, inputs):
        """Take the inputs to the network as dictionary and return a namedtuple
        of the input/index tensors to be embedded (GlyphTuple)"""
        embeddable_data = {}
        # Only flatten the data we want
        for key, value in inputs.items():
            if key in self.embeddings:
                # -- [ T x B x ...] -> [ B' x ... ]
                embeddable_data[key] = torch.flatten(value, 0, 1).long()

        # add our group id and subgroup id if we want them
        if self.requires_id_pairs_table:
            ids, groups = self.glyphs_to_idgroup(inputs["glyphs"])
            embeddable_data["groups"] = groups
            embeddable_data["subgroup_ids"] = ids

        # convert embeddable_data to a named tuple
        return self.GlyphTuple(**embeddable_data)

    def forward(self, data_tuple):
        """Output the embdedded tuple prepared in in prepare input. This will be
        a GlyphTuple."""
        embs = []
        for field, data in zip(self.targets, data_tuple):
            embs.append(self._select(self.embeddings[field], data))
        if len(embs) == 1:
            return embs[0]

        embedded = torch.cat(embs, dim=-1)
        if self.strategy.do_linear_layer:
            embedded = self.linear(embedded)
        return embedded

    def _select(self, embedding_layer, x):
        if self.use_index_select:
            out = embedding_layer.weight.index_select(0, x.view(-1))
            # handle reshaping x to 1-d and output back to N-d
            return out.view(x.shape + (-1,))
        else:
            return embedding_layer(x)

    def glyphs_to_idgroup(self, glyphs):
        T, B, H, W = glyphs.shape
        ids_groups = self.id_pairs_table.index_select(
            0, glyphs.view(-1).long()
        )
        ids = ids_groups.select(1, 0).view(T * B, H, W).long()
        groups = ids_groups.select(1, 1).view(T * B, H, W).long()
        return (ids, groups)
